import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
import spaic

from matplotlib.pyplot import *
global_list = list()
global_tv_dict = dict()
global_dv_dict = dict()
global_te_dict = dict()
global_de_dict = dict()

MIN_DV = 0.001
MAX_DV = 100.0

# global_v_dict[name]

class NeuronFunc(torch.autograd.Function):

    @staticmethod
    def forward(ctx, inputs_w: torch.Tensor, inputs_t: torch.Tensor, history_output: torch.Tensor,
                template_tv: torch.Tensor, template_dv: torch.Tensor, template_te: torch.Tensor, template_de: torch.Tensor,
                 threshold: torch.Tensor):
        # input shape : (batch_size, neuron_num, time_step)
        old_shape = list(inputs_t.shape)
        inputs = inputs_t.view(old_shape[0], -1, old_shape[-1])
        tmp_shape = inputs.shape

        temp_tv =  template_tv[:, :, :tmp_shape[2]]
        temp_dv = -template_dv[:, :, :tmp_shape[2]]
        temp_te =  template_te[:, :, :tmp_shape[2]]
        temp_de = -template_de[:, :, :tmp_shape[2]]
        current_tv = torch.matmul(inputs, temp_tv.unsqueeze(-1)) + torch.matmul(history_output, temp_te.unsqueeze(-1))
        current_dv = torch.matmul(inputs, temp_dv.unsqueeze(-1)) + torch.matmul(history_output, temp_de.unsqueeze(-1))


        current_spike = current_tv.gt(threshold).float()

        # for probabilistic spike
        # current_spike = torch.zeros_like(current_tv)
        # current_spike_mask = current_tv.gt(threshold)
        # current_spike.mask_fill_(torch.bitwise_not(current_spike_mask), F.relu(current_tv/threshold)/5.0)
        # current_spike.mask_fill_(current_spike_mask, 0.5*F.sigmoid(current_dv)+0.5)

        ctx.save_for_backward(inputs_w, history_output, current_spike, current_dv, temp_tv, temp_dv, temp_de)
        old_shape.append(1)

        return current_spike.view(old_shape)

    @staticmethod
    def backward(ctx, grad_t):

        inputs, history_output, current_spike, current_dv, partial_dv, partial_tv, partial_de = ctx.saved_tensors
        current_dv = torch.clamp(current_dv, MIN_DV, MAX_DV)

        old_shape = inputs.shape
        inputs = inputs.view(old_shape[0], -1, old_shape[-1])
        input_spike = torch.nonzero(inputs).float()

        partial_input_t = inputs         * partial_dv/current_dv
        partial_input_w = input_spike    * partial_tv/current_dv
        partial_history = history_output * partial_de/current_dv
        partial_input_t = partial_input_t.view(old_shape)*grad_t
        partial_input_w = partial_input_w.view(old_shape)*grad_t
        partial_history = partial_history.view(old_shape)*grad_t

        if ctx.needs_input_grad[3]:
            partial_tv = grad_t.view()*inputs / current_dv
            partial_te = history_output / current_dv
        else:
            partial_tv = partial_te = None




        return partial_input_w, partial_input_t, partial_history, partial_tv, None, partial_te, None, None

# func = UnitFunc.apply

class SpikePropUnit(nn.Module):

    def __init__(self, shape, name=None, threshold=1.0, n_step=1000, tau_m=10.0, tau_s=6.0,  dt=0.1, vt_curve=None, reset_curve=None, dv_curve=None, dreset_curve=None, istrain_neurparam=0):
        '''
        Initialize a Neruongroup with SpikeProp Learning rule.
        :param shape: shape of the neurongroup
        :param name: name of the neurongroup
        :param threshold:
        :param n_step:
        :param tau_m:
        :param tau_s:
        :param dt:
        :param vt_curve: user defined post-spike response potential (PSP)
        :param reset_curve:
        :param dv_curve:
        :param dreset_curve:
        :param train_neurparam: 0 -> don't train neural parameters, 1 -> train and use identical pararmeters for neuron group, 2 -> train and use different pararmeters for each neuron in the group
        NeuroFunc(inputs_w, inputs_t, history_output, template_tv, template_dv, template_te, template_de, threshold)
        '''

        super(SpikePropUnit, self).__init__()
        self.name = name
        self.shape = shape
        self.num = np.prod(shape, dtype=int)
        self.dt = dt
        self.n_step = n_step # max time step of grad backprop
        self.threshold = torch.tensor(threshold, dtype=torch.float)
        self.UserDefSRM = False
        self.op = NeuronFunc.apply
        self.istrain_neurparam = istrain_neurparam
        self.tau_m = tau_m
        self.tau_s = tau_s

        self.inputs_w_list = []
        self.inputs_t_list = []
        self.history_output_list = []

        if (vt_curve is not None) and (reset_curve is not None):
            # build user defined SRM model with input curve
            self.UserDefSRM = True
            if len(vt_curve) != n_step:
                raise ValueError("the lenght of user defined vt_curve is not accordant with n_step")
            if len(reset_curve) != n_step:
                raise ValueError("the lenght of user defined reset_curve is not accordant with n_step")
            tv = torch.tensor(vt_curve, dtype=torch.float)
            te = torch.tensor(reset_curve, dtype=torch.float)
            if dv_curve is None:
                dv = torch.zeros_like(tv)
                dv[1:] = (tv[1:] - tv[:-2])/dt
            elif len(dv_curve) != n_step:
                raise ValueError("the lenght of user defined dv_curve is not accordant with n_step")
            else:
                dv = torch.tensor(dv_curve, dtype=torch.float)
            if dreset_curve is None:
                de = torch.zeros_like(te)
                de[1:] = (te[1:]-te[:-2])/dt
            elif len(dreset_curve) != n_step:
                raise ValueError("the lenght of user defined dreset_curve is not accordant with n_step")
            else:
                de = torch.tensor(dreset_curve, dtype=torch.float)

            if istrain_neurparam == 0:
                self.tv = tv.view(1, 1, -1)
                self.dv = dv.view(1, 1, -1)
                self.te = te.view(1, 1, -1)
                self.de = de.view(1, 1, -1)
            elif istrain_neurparam == 1:
                self.tv = nn.Parameter(tv.view(1, 1, -1))
                self.dv = nn.Parameter(dv.view(1, 1, -1))
                self.te = nn.Parameter(te.view(1, 1, -1))
                self.de = nn.Parameter(de.view(1, 1, -1))
            elif istrain_neurparam == 2:
                self.tv = nn.Parameter(tv.view(1, 1, -1).repeat(1, self.num, 1))
                self.dv = nn.Parameter(dv.view(1, 1, -1).repeat(1, self.num, 1))
                self.te = nn.Parameter(te.view(1, 1, -1).repeat(1, self.num, 1))
                self.de = nn.Parameter(de.view(1, 1, -1).repeat(1, self.num, 1))
        else:
            # use default SRM model
            tt = torch.arange(1, n_step + 1, dtype=torch.float) * dt
            tt = tt.view(1, 1, -1)
            if istrain_neurparam == 0:
                self.tv = (torch.exp(-tt / self.tau_m) - torch.exp(-tt / self.tau_s)) / (self.tau_m - self.tau_s)
                self.dv = (torch.exp(-tt / self.tau_s) / self.tau_s - torch.exp(-tt / self.tau_m) / self.tau_m) / (self.tau_m - self.tau_s)
                self.te = -torch.exp(-tt / self.tau_m)
                self.de = torch.exp(-tt / self.tau_m) / self.tau_m
            elif istrain_neurparam == 1:
                self.tau_m = nn.Parameter(torch.tensor(self.tau_m))
                self.tau_s = nn.Parameter(torch.tensor(self.tau_s))
                self.tv = (torch.exp(-tt / self.tau_m) - torch.exp(-tt / self.tau_s)) / (self.tau_m - self.tau_s)
                self.dv = (torch.exp(-tt / self.tau_s) / self.tau_s - torch.exp(-tt / self.tau_m) / self.tau_m) / (self.tau_m - self.tau_s)
                self.te = -torch.exp(-tt / self.tau_m)
                self.de = torch.exp(-tt / self.tau_m) / self.tau_m
            elif istrain_neurparam == 2:
                self.tau_m = nn.Parameter(torch.tensor(self.tau_m).view(1, 1, 1).repeat(1, self.num, 1))
                self.tau_s = nn.Parameter(torch.tensor(self.tau_s).view(1, 1, 1).repeat(1, self.num, 1))
                self.tv = (torch.exp(-tt / self.tau_m) - torch.exp(-tt / self.tau_s)) / (self.tau_m - self.tau_s)
                self.dv = (torch.exp(-tt / self.tau_s) / self.tau_s - torch.exp(-tt / self.tau_m) / self.tau_m) / (self.tau_m - self.tau_s)
                self.te = -torch.exp(-tt / self.tau_m)
                self.de = torch.exp(-tt / self.tau_m) / self.tau_m

    def clear(self):
        self.inputs_w_list = []
        self.inputs_t_list = []
        self.history_output_list = []

    def forward(self, input_w, input_t):
        '''
        NeuroFunc(inputs_w, inputs_t, history_output, template_tv, template_dv, template_te, template_de, threshold)
        :param input_w:
        :param input_t:
        :return:
        '''

        self.inputs_w_list.append(input_w)
        self.inputs_t_list.append(input_t)
        if len(self.history_output_list) == 0:
            self.history_output_list.append(torch.zeros_like(input_w))
        inputs_w = torch.cat(self.inputs_w_list, dim=-1)
        inputs_t = torch.cat(self.inputs_t_list, dim=-1)
        history_outputs = torch.cat(self.history_output_list, dim=-1)
        output = self.op(inputs_w, inputs_t, history_outputs, self.tv, self.dv, self.te, self.de, self.threshold)
        self.history_output_list.append(output)
        return output


class SpkLinear(nn.Module):

    def __init__(self, input_n, output_n, weight_init=nn.init.xavier_uniform_):
        super(SpkLinear, self).__init__()
        self.weight = nn.Parameter(torch.Tensor(output_n, input_n))
        self.neuron = SpikePropUnit((output_n,))
        weight_init(self.weight)

    def forward(self, input):
        input_w = F.linear(input.detach(), self.weight)
        input_t = F.linear(input, self.weight.detach())
        output = self.neuron(input_w, input_t)
        return output

    def clear(self):
        self.neuron.clear()




class TestNet(nn.Module):

    def __init__(self):
        super(TestNet, self).__init__()
        self.layer1 = SpkLinear(784, 100)
        self.layer2 = SpkLinear(100, 10)

    def forward(self, spk):
        spk = self.layer1(spk)
        spk = self.layer2(spk)
        return spk

    def clear(self):
        self.layer1.clear()
        self.layer2.clear()


bat_size = 100
root = '../Datasets/MNIST'
train_set = spaic.MNIST(root, is_train=True)
test_set = spaic.MNIST(root, is_train=False)

# 创建DataLoader迭代器
train_loader = spaic.Dataloader(train_set, batch_size=bat_size, shuffle=True, drop_last=False)
test_loader = spaic.Dataloader(test_set, batch_size=bat_size, shuffle=False)
net = TestNet()
for item in train_loader:
    data, label = item
    inp = torch.tensor(data, dtype=torch.float)
    out = net(inp)
    print(out)





# -*- coding: utf-8 -*-
"""
Created on 2022/4/27
@project: SPAIC
@filename: Backprop-RSTDP
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Learning.Learner import Learner
from spaic.Network.BaseModule import Op
from spaic.Backend.Backend import Backend
import numpy as np
import torch


# a = torch.rand((2,1))
# b = torch.rand((1))
# print(torch.matmul(a,b))

class Backprop_RSTDP(Learner):
    _learner_count = 0
    
    def __init__(self, trainable=None, **kwargs):
        super(Backprop_RSTDP, self).__init__(trainable=trainable)
        self.prefered_backend = ['pytorch']
        self.name = 'Backprop-RSTDP' + str(Backprop_RSTDP._learner_count)
        Backprop_RSTDP._learner_count += 1
        self._tau_constant_variables = dict()
        self._constant_variables = dict()
        self._variables = dict()

        self.learning_rate = kwargs.get('lr', 0.1)
        self.homeostatic_rate = kwargs.get('hs_r',0.01)
        self.reward_var_names = kwargs.get('reward_vars', 'Output_Reward')
        self.reward_weight_min = kwargs.get('reward_wegith_min', -10.0)
        self.reward_weight_max = kwargs.get('reward_wegith_max', 10.0)

        self._tau_constant_variables['tau_pre'] = kwargs.get('tau_pre', 5.0)
        self._tau_constant_variables['tau_post'] = kwargs.get('tau_post', 12.0)
        self._constant_variables['A_plus'] = kwargs.get('A_plus', 1.0)
        self._constant_variables['A_minus'] = kwargs.get('A_minus', -0.4)
        self.traces = []
        self.spks = []


    def spk_trace_update(self, trace, beta, spk):
        with torch.no_grad():
            trace = beta*trace + (1-beta)*spk
            # if list(trace.shape) == [200,400]:
            #     self.traces.append(trace)
            #     self.spks.append(spk)
        return trace


    # def RSTDP_update(self, trace_pre, trace_post, beta_pre, beta_post, spk_pre, spk_post, A_plus, A_minus):
    #     with torch.no_grad():
    #         trace_pre = beta_pre*trace_pre + spk_pre
    #         trace_post = beta_post*trace_post + spk_post
    #         if list(trace_pre.shape) == [200,784]:
    #             self.traces.append(trace_post)
    #             self.spks.append(spk_post)
    #         eligibility = torch.mean(A_plus*trace_pre.unsqueeze(1)*spk_post.unsqueeze(2) + A_minus*trace_post.unsqueeze(2)*spk_pre.unsqueeze(1), 0)
    #         homeo_eligibility = torch.mean(trace_pre.unsqueeze(1)*(trace_post**2).unsqueeze(2), 0)
    #     return trace_pre, trace_post, eligibility, homeo_eligibility

    def RSTDP_update(self, weight, trace_pre, trace_post,  spk_pre, spk_post, A_plus, A_minus, reward_weight, reward):

        with torch.no_grad():
            reward = torch.matmul(reward, reward_weight.t()) # shape of post neuron (pre of reward_weight)
            reward = reward.unsqueeze(2)

            eligibility = torch.mean(A_plus*torch.matmul(spk_post.unsqueeze(2), trace_pre.unsqueeze(1)) + A_minus*torch.matmul(trace_post.unsqueeze(2), spk_pre.unsqueeze(1)), 0)
            homeo_eligibility = torch.mean(torch.matmul((trace_post - 0.1).unsqueeze(2),trace_pre.unsqueeze(1)), 0)

            delta_w = torch.mean(self.learning_rate*reward*eligibility - self.homeostatic_rate*homeo_eligibility,0)
            weight.add_(delta_w)
        return weight

    def build(self, backend: Backend):
        super(Backprop_RSTDP, self).build(backend)
        self.dt = backend.dt

        for (key, tau_var) in self._tau_constant_variables.items():
            tau_var = np.exp(-self.dt / tau_var)
            shape = ()
            self.variable_to_backend(self.get_varname(key), shape, value=tau_var)

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(self.get_varname(key), shape, value=var)

        if isinstance(self.reward_var_names, str):
            single_reward = True
        else:
            assert len(self.reward_var_names) == len(self.trainable_connections)
            single_reward = False

        for ind, conn in enumerate(self.trainable_connections.values()):
            preg = conn.pre
            postg = conn.post
            pre_name = conn.get_pre_name(preg, 'O')
            post_name = conn.get_post_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')

            trace_pre_name = preg.get_labeled_name('spk_trace')
            if not backend.has_variable(trace_pre_name):
                self.variable_to_backend(trace_pre_name, backend.get_varialble(pre_name).shape, value=0.0)
                self.op_to_backend([trace_pre_name], self.spk_trace_update, [trace_pre_name, self.get_varname('tau_pre'), pre_name])
            trace_post_name = postg.get_labeled_name('spk_trace')
            if not backend.has_variable(trace_post_name):
                self.variable_to_backend(trace_post_name, backend.get_varialble(post_name).shape, value=0.0)
                self.op_to_backend([trace_post_name], self.spk_trace_update,
                                       [trace_post_name, self.get_varname('tau_post'), post_name])
            # eligibility_name = conn.get_link_name(preg, postg, 'eligibility')
            # homeo_eli_name = conn.get_link_name(preg, postg, 'homeostatic_eligibility')
            # self.variable_to_backend(eligibility_name, backend.get_varialble(weight_name).shape, value=0.0)
            # self.variable_to_backend(homeo_eli_name, backend.get_varialble(weight_name).shape, value=0.0)
            # self.op_to_backend([trace_pre_name, trace_post_name, eligibility_name, homeo_eli_name), self.RSTDP_update,
            #                        [trace_pre_name, trace_post_name, self.get_varname('tau_pre'), self.get_varname('tau_post'),
            #                         pre_name, post_name, self.get_varname('A_plus'), self.get_varname('A_minus')]])

            if single_reward:
                reward_name = self.reward_var_names
                reward_shape = backend.get_varialble(reward_name).shape
            else:
                reward_name = self.reward_var_names[ind]
                reward_shape = backend.get_varialble(reward_name).shape
            post_shape = backend.get_varialble(post_name).shape
            assert len(reward_shape) == 2 and len(post_shape) == 2
            reward_weight_shape = [post_shape[1], reward_shape[1]]

            if reward_weight_shape == [10,10]:
                reward_weight_value = np.eye(10)
            else:
                reward_weight_value = np.clip(0.2*np.random.randn(*reward_weight_shape), self.reward_weight_min, self.reward_weight_max)
            reward_weight_name = conn.get_link_name(preg, postg, 'reward_weight')

            self.variable_to_backend(reward_weight_name, reward_weight_shape, reward_weight_value)
            self.op_to_backend([weight_name], self.RSTDP_update, [weight_name, trace_pre_name, trace_post_name, pre_name,
                                                                       post_name, self.get_varname('A_plus'), self.get_varname('A_minus'),
                                                                       reward_weight_name, reward_name])

Learner.register('bp_rstdp', Backprop_RSTDP)













class Backprop_RSTDPET(Learner):
    _learner_count = 0

    def __init__(self, trainable=None, **kwargs):
        super(Backprop_RSTDPET, self).__init__(trainable=trainable)
        self.prefered_backend = ['pytorch']
        self.name = 'Backprop-RSTDP' + str(Backprop_RSTDPET._learner_count)
        Backprop_RSTDPET._learner_count += 1
        self._tau_constant_variables = dict()
        self._constant_variables = dict()
        self._variables = dict()

        self.learning_rate = kwargs.get('lr', 0.1)
        self.homeostatic_rate = kwargs.get('hs_r', 0.01)
        self.reward_var_names = kwargs.get('reward_vars', 'Output_Reward')
        self.reward_weight_min = kwargs.get('reward_wegith_min', -10.0)
        self.reward_weight_max = kwargs.get('reward_wegith_max', 10.0)

        self._tau_constant_variables['tau_pre'] = kwargs.get('tau_pre', 4.0)
        self._tau_constant_variables['tau_post'] = kwargs.get('tau_post', 10.0)
        self._tau_constant_variables['tau_et'] = kwargs.get('tau_et', 5.0)
        self._constant_variables['A_plus'] = kwargs.get('A_plus', 1.0)
        self._constant_variables['A_minus'] = kwargs.get('A_minus', -0.5)
        self.traces = []
        self.spks = []

    def spk_trace_update(self, trace, beta, spk):
        with torch.no_grad():
            trace = beta * trace + (1 - beta) * spk
            # if list(trace.shape) == [200,400]:
            #     self.traces.append(trace)
            #     self.spks.append(spk)
        return trace

    # def RSTDP_update(self, trace_pre, trace_post, beta_pre, beta_post, spk_pre, spk_post, A_plus, A_minus):
    #     with torch.no_grad():
    #         trace_pre = beta_pre*trace_pre + spk_pre
    #         trace_post = beta_post*trace_post + spk_post
    #         if list(trace_pre.shape) == [200,784]:
    #             self.traces.append(trace_post)
    #             self.spks.append(spk_post)
    #         eligibility = torch.mean(A_plus*trace_pre.unsqueeze(1)*spk_post.unsqueeze(2) + A_minus*trace_post.unsqueeze(2)*spk_pre.unsqueeze(1), 0)
    #         homeo_eligibility = torch.mean(trace_pre.unsqueeze(1)*(trace_post**2).unsqueeze(2), 0)
    #     return trace_pre, trace_post, eligibility, homeo_eligibility

    def RSTDPET_update(self, weight, et, trace_pre, trace_post, spk_pre, spk_post, tau_et, A_plus, A_minus, reward_weight, reward):

        with torch.no_grad():
            reward = torch.matmul(reward, reward_weight.t())  # shape of post neuron (pre of reward_weight)
            reward = reward.unsqueeze(2)

            eligibility = torch.mean(
                A_plus * torch.matmul(spk_post.unsqueeze(2), trace_pre.unsqueeze(1)),0)
            et = et*tau_et + eligibility
            homeo_eligibility =  torch.matmul(trace_post.unsqueeze(2), spk_pre.unsqueeze(1))*(trace_post-0.2).unsqueeze(2)

            delta_w = torch.mean(self.learning_rate * reward * et - self.homeostatic_rate * homeo_eligibility, 0)
            weight.add_(delta_w)
        return et

    def build(self, backend: Backend):
        super(Backprop_RSTDPET, self).build(backend)
        self.dt = backend.dt

        for (key, tau_var) in self._tau_constant_variables.items():
            tau_var = np.exp(-self.dt / tau_var)
            shape = ()
            self.variable_to_backend(self.get_varname(key), shape, value=tau_var)

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(self.get_varname(key), shape, value=var)

        if isinstance(self.reward_var_names, str):
            single_reward = True
        else:
            assert len(self.reward_var_names) == len(self.trainable_connections)
            single_reward = False

        for ind, conn in enumerate(self.trainable_connections.values()):
            preg = conn.pre
            postg = conn.post
            pre_name = conn.get_pre_name(preg, 'O')
            post_name = conn.get_post_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')


            trace_pre_name = preg.get_labeled_name('spk_trace')
            if not backend.has_variable(trace_pre_name):
                self.variable_to_backend(trace_pre_name, backend.get_varialble(pre_name).shape, value=0.0)
                self.op_to_backend([trace_pre_name], self.spk_trace_update,
                                       [trace_pre_name, self.get_varname('tau_pre'), pre_name])
            trace_post_name = postg.get_labeled_name('spk_trace')
            if not backend.has_variable(trace_post_name):
                self.variable_to_backend(trace_post_name, backend.get_varialble(post_name).shape, value=0.0)
                self.op_to_backend([trace_post_name], self.spk_trace_update,
                                       [trace_post_name, self.get_varname('tau_post'), post_name])
            # eligibility_name = conn.get_link_name(preg, postg, 'eligibility')
            # homeo_eli_name = conn.get_link_name(preg, postg, 'homeostatic_eligibility')
            eligibility_trace_name = conn.get_link_name(preg, postg, 'et')
            self.variable_to_backend(eligibility_trace_name, backend.get_varialble(weight_name).shape, value=0.0)
            # self.variable_to_backend(homeo_eli_name, backend.get_varialble(weight_name).shape, value=0.0)
            # self.op_to_backend([trace_pre_name, trace_post_name, eligibility_name, homeo_eli_name), self.RSTDP_update,
            #                        [trace_pre_name, trace_post_name, self.get_varname('tau_pre'), self.get_varname('tau_post'),
            #                         pre_name, post_name, self.get_varname('A_plus'), self.get_varname('A_minus')]])

            if single_reward:
                reward_name = self.reward_var_names
                reward_shape = backend.get_varialble(reward_name).shape
            else:
                reward_name = self.reward_var_names[ind]
                reward_shape = backend.get_varialble(reward_name).shape
            post_shape = backend.get_varialble(post_name).shape
            assert len(reward_shape) == 2 and len(post_shape) == 2
            reward_weight_shape = [post_shape[1], reward_shape[1]]

            if reward_weight_shape == [10, 10]:
                reward_weight_value = np.eye(10)
            else:
                reward_weight_value = np.clip(np.random.randn(*reward_weight_shape), self.reward_weight_min,
                                              self.reward_weight_max)
            reward_weight_name = conn.get_link_name(preg, postg, 'reward_weight')

            self.variable_to_backend(reward_weight_name, reward_weight_shape, reward_weight_value)
            self.op_to_backend(
                ([eligibility_trace_name], self.RSTDPET_update, [weight_name, eligibility_trace_name, trace_pre_name,
                                                                 trace_post_name, pre_name, post_name,
                                                                 self.get_varname('tau_et'), self.get_varname('A_plus'),
                                                      self.get_varname('A_minus'), reward_weight_name, reward_name]))


Learner.register('bp_rstdpet', Backprop_RSTDP)






# -*- coding: utf-8 -*-
"""
Created on 2020/8/11
@project: SPAIC
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""

from .Network import Network
from .Topology import Projection, Connection
from .Assembly import Assembly
from .Connections import FullConnection
from .Synapse import Basic_synapse



# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Rewards.py
@time:2021/12/2 9:43
@description:
"""
from .Node import Node, Reward
import torch
import numpy as np

class Global_Reward(Reward):

    def __init__(self,shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Global_Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.pop_size = kwargs.get('pop_size', 1)
        self.reward_signal = kwargs.get('reward_signal', 1)
        self.punish_signal = kwargs.get('punish_signal', -1)

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        spike_rate = record.sum(0)
        pop_num = int(self.num / self.pop_size)
        pop_spikes_temp = (
            [
                spike_rate[:, (i * self.pop_size): (i * self.pop_size) + self.pop_size].sum(dim=1)
                for i in range(pop_num)
            ]
        )
        pop_spikes = torch.stack(pop_spikes_temp, dim=-1)
        predict = torch.argmax(pop_spikes, dim=1)  # return the indices of the maximum values of a tensor across columns.
        reward = self.punish_signal * torch.ones(predict.shape, device=device)
        flag = torch.tensor([predict[i] == target[i] for i in range(predict.size(0))])
        reward[flag] = self.reward_signal
        if len(reward) > 1:
            reward = reward.mean()
        return reward

Reward.register('global_reward', Global_Reward)

class DA_Reward(Reward):

    def __init__(self,shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(DA_Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.pop_size = kwargs.get('pop_size', 1)
        self.reward_signal = kwargs.get('reward_signal', 1)
        self.punish_signal = kwargs.get('punish_signal', 0)



    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        spike_rate = record.sum(0)
        pop_num = int(self.num / self.pop_size)
        pop_spikes_temp = (
            [
                spike_rate[:, (i * self.pop_size): (i * self.pop_size) + self.pop_size].sum(dim=1)
                for i in range(pop_num)
            ]
        )
        pop_spikes = torch.stack(pop_spikes_temp, dim=-1)
        predict = torch.argmax(pop_spikes, dim=1)  # return the indices of the maximum values of a tensor across columns.
        reward = self.punish_signal * torch.ones(spike_rate.shape, device=device)
        for i in range(len(target)):
            if predict[i] == target[i]:
                reward[i, predict[i]*self.pop_size:(predict[i]+1)*self.pop_size] = self.reward_signal
        if reward.size(0) > 1:
            reward = reward.sum(dim=0).unsqueeze(dim=0)
        return reward

Reward.register('da_reward', DA_Reward)

class XOR_Reward(Reward):


    def __init__(self,shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(XOR_Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.pop_size = kwargs.get('pop_size', 1)
        self.reward_signal = kwargs.get('reward_signal', 1)
        self.punish_signal = kwargs.get('punish_signal', -1)

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        spike_rate = record.sum(0)
        pop_num = int(self.num / self.pop_size)
        pop_spikes_temp = (
            [
                spike_rate[:, (i * self.pop_size): (i * self.pop_size) + self.pop_size].sum(dim=1)
                for i in range(pop_num)
            ]
        )
        pop_spikes = torch.stack(pop_spikes_temp, dim=-1)
        reward = self.punish_signal * torch.ones(pop_spikes.shape, device=device)
        if target == 1:
            if pop_spikes > 0:
                reward = torch.tensor(self.reward_signal, device=device)

        elif target == 0:
            if pop_spikes > 0:
                reward = torch.tensor(self.punish_signal, device=device)
        return reward

Reward.register('xor_reward', XOR_Reward)

class Environment_Reward(Reward):


    def __init__(self,shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Environment_Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        reward = torch.tensor(target, device=device, dtype=self.data_type)
        return reward

Reward.register('environment_reward', Environment_Reward)


class Classifiy_Reward(Reward):

    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Classifiy_Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type,
                                                 **kwargs)
        self.beta = 0.99
        self.out = 0
        # self.pos_reward = 0
        # self.neg_reward = 0
        self.rewards = []

    def init_state(self):
        super(Classifiy_Reward, self).init_state()
        self.out = 0
        # self.pos_reward = 0
        # self.neg_reward = 0
        self.rewards = []

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        target_index = torch.tensor(target, device=device).view(-1,1)
        output = torch.mean(record, 0)
        mask = torch.zeros_like(output)
        mask.scatter_(1, target_index, 1)
        self.out = self.beta*self.out + output
        output = (self.out - torch.mean(self.out ).detach()) / (torch.std(self.out).detach() + 0.1)
        softmax = torch.softmax(output+1.0e-8, 1)

        # self.pos_reward = self.beta*self.pos_reward + mask*output
        # self.neg_reward = self.beta*self.neg_reward + (1-mask)*output

        reward = (1-softmax)*mask - softmax*(1-mask)
        # print(torch.mean(reward))
        # self.rewards.append(reward)
        return reward


Reward.register('classify_reward', Classifiy_Reward)


from .Learner import Learner
import torch
import numpy as np

class Base_STDP(Learner):
    def __init__(self, trainable=None, *args, **kwargs):
        super(Base_STDP, self).__init__(trainable, *args, **kwargs)
        self.gradient_based = False
    
    def build(self, backend):
        super(Base_STDP, self).build(backend)


class nearest_online_STDP(Base_STDP):
    '''
        nearest_online STDP learning rule.

        Args:
            Apost(num) : The parameter Apost of nearest_online STDP learning model.
            Apre(num) : The parameter Apre of nearest_online STDP learning model.
            trace_decay(num) : The parameter trace_decay of nearest_online STDP learning model.
            preferred_backend(list) : The backend prefer to use, should be a list.
            name(str) : The name of this learning model. Should be 'nearest_online_STDP'.

        Methods:
            initial_param(self, input, output): initialize the output_trace and the input_trace for each batch.
            nearest_online_stdp_weightupdate(self, input, output, weight): calculate the update of weight
            build(self, backend): Build the backend, realize the algorithm of nearest_online STDP learning model.

        Example:
            self._learner = BaseLearner(algorithm='nearest_online_STDP', lr=0.5, trainable=self, conn=self.connection1)

        Reference:
            Unsupervised learning of digit recognition using spike-timing-dependent plasticity.
            doi: 10.3389/fncom.2015.00099.
            url: http://journal.frontiersin.org/article/10.3389/fncom.2015.00099/abstract
    '''

    def __init__(self, trainable=None, *args, **kwargs):
        super(nearest_online_STDP, self).__init__(trainable=trainable)

        self.prefered_backend = ['pytorch']
        self.name = 'nearest_online_STDP'
        self._constant_variables = dict()
        self._constant_variables['Apost'] = kwargs.get('Apost', 1e-2)
        self._constant_variables['Apre'] = kwargs.get('Apre', 1e-4)
        self._constant_variables['trace_pre'] = kwargs.get('trace_pre', np.exp(-1/20))
        self._constant_variables['trace_post'] = kwargs.get('trace_post', np.exp(-1/20))
        self._constant_variables['spike'] = kwargs.get('spike', 1)
        self.lr = kwargs.get('lr', 0.01)
        self.w_norm = 78.4
        self.trainable = trainable

    def nearest_online_stdp_weightupdate(self, dw, weight):
        '''

            Args:
                dw: the change of weight
                weight: weight between pre and post neurongroup

            Returns:
                Updated weight.

        '''
        with torch.no_grad():
            # weight = weight + dw
            weight.add_(dw)

            # if self._backend.n_time_step < self.total_step:
            #     pass
            # else:
            #     for i in range(weight.shape[0]):
            #         weight[i] = (weight[i] * self.w_norm) / torch.sum(torch.abs(weight[i]))

            weight[...] = (self.w_norm * torch.div(weight, torch.sum(torch.abs(weight), 1, keepdim=True)))
            weight.clamp_(0.0, 1.0)
            return weight


    def build(self, backend):

        super(nearest_online_STDP, self).build(backend)

        self._backend = backend
        self.dt = backend.dt
        self.run_time = backend.runtime
        self.total_step = self.run_time / self.dt - 1

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(key, shape, value=var)

        for conn in self.trainable_connections.values():
            preg = conn.pre
            postg = conn.post
            # pre_name = conn.get_input_name(preg, postg)
            # post_name = conn.get_target_output_name(postg)
            # weight_name = conn.get_weight_name(preg, postg)
            pre_name = conn.get_input_name(preg, postg)
            post_name = conn.get_group_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')

            input_trace_name = pre_name + '_{input_trace}'
            output_trace_name = post_name + '_{output_trace}'
            dw_name = weight_name + '_{dw}'
            self.variable_to_backend(input_trace_name, backend._variables[pre_name].shape, value=0.0)
            self.variable_to_backend(output_trace_name, backend._variables[post_name].shape, value=0.0)
            self.variable_to_backend(dw_name, backend._variables[weight_name].shape, value=0.0)
            # input_trace_s = (self.input_trace * self.trace_decay)
            # input_trace_s = input + (1 - input) * input_trace_s
            # self.input_trace = input_trace_s
            #
            # output_trace_s = (self.output_trace * self.trace_decay)
            # output_trace_s = output + (1 - output) * output_trace_s
            # self.output_trace = output_trace_s
            #
            ##calculate dw, dw = Apost * (output_spike * input_trace) – Apre * (output_trace * input_spike)
            #
            # dw = self.Apost * torch.matmul(output_spike.permute(1, 0), input_trace_s) \
            #              - self.Apre * torch.matmul(output_trace_s.permute(1, 0), input_spike)  #

            self.op_to_backend('input_trace_s', 'var_mult', [input_trace_name, 'trace_pre'])
            self.op_to_backend('input_temp', 'minus', ['spike', pre_name])
            self.op_to_backend(input_trace_name, 'var_linear', ['input_temp', 'input_trace_s', pre_name])

            self.op_to_backend('output_trace_s', 'var_mult', [output_trace_name, 'trace_post'])
            self.op_to_backend('output_temp', 'minus', ['spike', post_name])
            self.op_to_backend(output_trace_name, 'var_linear', ['output_temp', 'output_trace_s', post_name])

            self.op_to_backend('pre_post_temp', 'mat_mult_pre', [post_name, input_trace_name+'[updated]'])
            self.op_to_backend('pre_post', 'var_mult', ['Apost', 'pre_post_temp'])
            self.op_to_backend('post_pre_temp', 'mat_mult_pre', [output_trace_name+'[updated]', pre_name])
            self.op_to_backend('post_pre', 'var_mult', ['Apre', 'post_pre_temp'])
            self.op_to_backend(dw_name, 'minus', ['pre_post', 'post_pre'])
            self.op_to_backend(weight_name, self.nearest_online_stdp_weightupdate, [dw_name, weight_name])
Learner.register("nearest_online_stdp", nearest_online_STDP)



class full_online_STDP(Base_STDP):
    '''
        nearest_online STDP learning rule.

        Args:
            Apost(num) : The parameter Apost of full_online STDP learning model.
            Apre(num) : The parameter Apre of full_online STDP learning model.
            trace_decay(num) : The parameter trace_decay of full_online STDP learning model.
            preferred_backend(list) : The backend prefer to use, should be a list.
            name(str) : The name of this learning model. Should be 'full_online_STDP'.

        Methods:
            initial_param(self, input, output): initialize the output_trace and the input_trace for each batch.
            nearest_online_stdp_weightupdate(self, input, output, weight): calculate the update of weight
            build(self, backend): Build the backend, realize the algorithm of full_online STDP learning model.

        Example:
            self._learner = BaseLearner(algorithm='full_online_STDP', lr=0.5, trainable=self, conn=self.connection1)

        Reference:
            Unsupervised learning of digit recognition using spike-timing-dependent plasticity.
            doi: 10.3389/fncom.2015.00099.
            url: http://journal.frontiersin.org/article/10.3389/fncom.2015.00099/abstract
        '''

    def __init__(self, trainable=None, *args, **kwargs):
        super(full_online_STDP, self).__init__(trainable=trainable)
        self.prefered_backend = ['pytorch']
        self.firing_func = None
        self._constant_variables = dict()
        self._constant_variables['Apost'] = kwargs.get('Apost', 0.01)
        self._constant_variables['Apre'] = kwargs.get('Apre', 1e-4)
        self._constant_variables['trace_decay'] = kwargs.get('trace_decay', np.exp(-1 / 20))
        self.lr = kwargs.get('lr', 0.01)
        self.name = 'full_online_STDP'
        self.w_norm = 78.4

    def full_online_stdp_weightupdate(self, dw, weight):
        '''

            Args:
                dw: the change of weight
                weight: weight between pre and post neurongroup

            Returns:
                Updated weight.

        '''
        with torch.no_grad():
            weight.add_(dw)

            if self._backend.n_time_step < self.total_step:
                pass
            else:
                weight[...] = (self.w_norm * torch.div(weight, torch.sum(torch.abs(weight), 1, keepdim=True)))

            weight.clamp_(0.0, 1.0)
            return weight


    def build(self, backend):
        super(full_online_STDP, self).build(backend)
        self._backend = backend
        self.dt = backend.dt
        self.run_time = backend.runtime
        self.total_step = self.run_time / self.dt - 1

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(key, shape, value=var)

        for conn in self.trainable_connections.values():
            preg = conn.pre
            postg = conn.post
            pre_name = conn.get_input_name(preg, postg)
            post_name = conn.get_group_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')

            input_trace_name = pre_name + '_{input_trace}'
            output_trace_name = post_name + '_{output_trace}'
            dw_name = weight_name + '_{dw}'
            self.variable_to_backend(input_trace_name, backend._variables[pre_name].shape, value=0.0)
            self.variable_to_backend(output_trace_name, backend._variables[post_name].shape, value=0.0)
            self.variable_to_backend(dw_name, backend._variables[weight_name].shape, value=0.0)

            # input_trace_s = (self.input_trace * self.trace_decay) + input
            # self.input_trace = input_trace_s
            #
            # output_trace_s = (self.output_trace * self.trace_decay) + output
            # self.output_trace = output_trace_s
            #
            #calculate dw, dw = Apost * (output_spike * input_trace) – Apre * (output_trace * input_spike)
            #
            # dw = self.Apost * torch.matmul(output_spike.permute(1, 0), input_trace_s) \
            #              - self.Apre * torch.matmul(output_trace_s.permute(1, 0), input_spike)  #

            self.op_to_backend('input_trace_temp', 'var_mult', [input_trace_name, 'trace_decay'])
            self.op_to_backend(input_trace_name, 'add', [pre_name, 'input_trace_temp'])


            self.op_to_backend('output_trace_temp', 'var_mult', [output_trace_name, 'trace_decay'])
            self.op_to_backend(output_trace_name, 'add', [post_name, 'output_trace_temp'])

            self.op_to_backend('pre_post_temp', 'mat_mult_pre', [post_name, input_trace_name+'[updated]'])
            self.op_to_backend('pre_post', 'var_mult', ['Apost', 'pre_post_temp'])
            self.op_to_backend('post_pre_temp', 'mat_mult_pre', [output_trace_name+'[updated]', pre_name])
            self.op_to_backend('post_pre', 'var_mult', ['Apre', 'post_pre_temp'])
            self.op_to_backend(dw_name, 'minus', ['pre_post', 'post_pre'])
            self.op_to_backend(weight_name, self.full_online_stdp_weightupdate,[dw_name, weight_name])



Learner.register("full_online_STDP", full_online_STDP)


class Meta_nearest_online_STDP(Base_STDP):
    def __init__(self, trainable=None, *args, **kwargs):
        super(Meta_nearest_online_STDP, self).__init__(trainable=trainable, **kwargs)
        self.trainable = trainable
        self.prefered_backend = ['pytorch']
        self.name = 'meta_nearest_online_STDP'
        self._constant_variables = dict()
        self._constant_variables['Apost'] = kwargs.get('Apost', 1e-2)
        self._constant_variables['Apre'] = kwargs.get('Apre', 1e-4)
        self._constant_variables['pre_decay'] = kwargs.get('pre_decay', np.exp(-1/30))
        self._constant_variables['post_decay'] = kwargs.get('post_decay', np.exp(-1 / 200))
        self.w_min = kwargs.get('w_min', 0.0)
        self.w_max = kwargs.get('w_max', 1.0)
        self.w_norm = 1.2
        self.w_mean = None
        self.lr = kwargs.get('lr', 0.01)
        self.param_run_update = True

    def update(self, input, output, input_trace, output_trace, pre_decay, post_decay, Apost, Apre, weight):
        if self.w_mean is None:
            self.w_mean = torch.mean(weight, dim=1, keepdim=True).detach()
            self.aw_mean = torch.mean(self.w_mean)
        if self.training:
            input_trace = pre_decay*input_trace*input.le(0) + input
            output_trace = post_decay*output_trace*output.le(0) + output
            pre_post = torch.matmul(output.permute(1, 0), input_trace)
            post_pre = torch.matmul(output_trace.permute(1,0), input)
            dw = Apost*pre_post - Apre*post_pre
            with torch.no_grad():
                self.w_mean = self.w_mean - 0.1*Apre * (
                            torch.mean(output_trace, 0) - torch.mean(output_trace)).unsqueeze(1)
                self.w_mean = self.w_mean*self.aw_mean/torch.mean(self.w_mean)
            soft_clamp = (dw.lt(0)*torch.exp(-0.2/torch.clamp_min(weight-self.w_min, 1.0e-4))
                          + dw.gt(0)*torch.exp(-0.2/torch.clamp_min(self.w_max-weight, 1.0e-4))).detach()
            weight = weight*self.w_mean/(torch.clamp_min(torch.mean(weight, dim=1, keepdim=True), 1e-6)).detach() + dw*soft_clamp
        # mw = torch.mean(weight, dim=1)
        # print(torch.amax(mw), torch.amin(mw))
        return input_trace, output_trace, weight

    def build(self, backend):

        super(Meta_nearest_online_STDP, self).build(backend)

        self._backend = backend
        self.dt = backend.dt
        self.run_time = backend.runtime

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(key, shape, value=var)

        for conn in self.trainable_connections.values():
            preg = conn.pre
            postg = conn.post
            pre_name = conn.get_input_name(preg, postg)
            post_name = conn.get_group_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')

            input_trace_name = pre_name + '_{input_trace}'
            output_trace_name = post_name + '_{output_trace[stay]}'
            self.variable_to_backend(input_trace_name, backend._variables[pre_name].shape, value=0.0)
            self.variable_to_backend(output_trace_name, backend._variables[post_name].shape, value=0.0)

            self.op_to_backend([input_trace_name, output_trace_name, weight_name], self.update,
                                   [pre_name, post_name, input_trace_name, output_trace_name, 'pre_decay', 'post_decay', 'Apost', 'Apre', weight_name])

Learner.register("meta_nearest_online_stdp", Meta_nearest_online_STDP)



# -*- coding: utf-8 -*-
"""
Created on 2020/8/11
@project: SPAIC
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""

VERSION = (0, 6, 2, 0, 0)
__version__ = '.'.join(map(lambda x: str(x), VERSION))

# print('__file__={0:<35} | __name__={1:<20} | __package__={2:<20}'.format(__file__,__name__,str(__package__)))
from .Network import Network, Connection, Assembly, Projection, Synapse
from .Neuron import NeuronGroup, NeuronModel
from .Neuron import Node, Encoders, Decoders, Generators, Rewards, Actions
from .Neuron.Node import Encoder, Decoder, Generator, Reward, Action
from .Neuron.Module import Module
from .Network.BaseModule import BaseModule, Op
from .Network.ConnectPolicy import ExcludedTypePolicy, IndexConnectPolicy, IncludedTypePolicy
from .Backend.Backend import Backend
from .Backend.Torch_Backend import Torch_Backend
# from .Backend.Tensorflow_Backend import Tensorflow_Backend
from .Monitor.Monitor import StateMonitor, SpikeMonitor
from .Learning.Rate_Modulation import Rate_Modulate
from .Learning.STCA_Learner import STCA
from .Learning.SurrogateBP_Learner import SurrogateBP

from .Learning.Learner import Learner
from .IO.Dataset import Dataset, CustomDataset, MNIST, FashionMNIST, OctMNIST, PathMNIST, MNISTVoices, cifar10, SHD, SSC, \
    DVS128Gesture
from .IO.Initializer import BaseInitializer, uniform, normal, xavier_normal, xavier_uniform, kaiming_normal,\
    kaiming_uniform, constant, sparse
from .IO.Dataloader import Dataloader
from .IO.Pipeline import RLPipeline, ReplayMemory
from .IO.Environment import GymEnvironment
import numpy as np

from .Library import Network_loader, Network_saver



# ============== Global variable block for network building ==================
global_assembly_context_list = list()
global_assembly_context_omit_start = 10000000000000000000000
global_assembly_context_omit_end = -1
global_assembly_init_count = 0
global_module_name_count = 0
debug_grad = dict()




# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Visualization
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Network.Assembly import NetModule

class Visualization(NetModule):

    def __init__(self):
        super(Visualization, self).__init__()



# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Actions.py
@time:2021/12/3 9:22
@description:
"""
from .Node import Action
import torch
import numpy as np
class PopulationRate_Action(Action):
    """
    Selects an action probabilistically based on output spiking activity of population.
    """
    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(PopulationRate_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        """
        Args:
            record: spiking activity of output layer. The shape of record is (time_step, batch_size, n_neurons)
        Returns:
            Action sampled from population output activity.
        """

        assert record.shape[2] % self.num == 0, (
            f"Output layer size of {record.shape[2]} is not divisible by action space size of"
            f" {self.num}."
        )

        pop_size = int(record.shape[2] / self.num)
        spike_num = record.sum().type(self._backend.data_type)

        # Choose action based on population's spiking.
        if spike_num == 0:
            action = torch.randint(low=0, high=self.num, size=(1,))[0]
        else:
            pop_spikes = torch.tensor(
                [
                    record[:, :, (i * pop_size): (i * pop_size) + pop_size].sum()
                    for i in range(self.num)
                ],
                device=record.device,
            ).type(self._backend.data_type)
            # multinomial(input, num_samples,replacement=False), 采样的时候是根据输入张量的数值当做权重来进行抽样的, 返回index, 数值越大, 抽到的可能性越大, 如果是0 则不会抽到
            action = torch.multinomial((pop_spikes / spike_num).view(-1), 1)[0].item()
        return action

Action.register('pop_rate_action', PopulationRate_Action)

class Softmax_Action(Action):

    """
    Selects an action using softmax function based on spiking of output layer.
    """

    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Softmax_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        """
        Args:
            record: spiking activity of output layer.
        Returns:
            Action sampled from softmax over spiking activity of output layer.
        """
        assert (
            record.shape[2] == self.num
        ), "Output layer size is not equal to the size of the action space."
        spikes = torch.sum(record, dim=0)
        probabilities = torch.softmax(spikes, dim=0)
        return torch.multinomial(probabilities, num_samples=1).item()

Action.register('softmax_action', Softmax_Action)


class Highest_Spikes_Action(Action):

    """
    Selects an action that has the highest firing rate. In case of equal spiking select randomly
    """

    def __init__(self,shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Highest_Spikes_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        """
        Args:
            record: spiking activity of output layer.
        Returns:
            Action sampled from highest activities of output layer.
        """

        assert (
           record.shape[2] == self.num
        ), "Output layer size is not equal to the size of the action space."


        spikes = torch.sum(record, dim=0).squeeze()
        action = torch.where(spikes == spikes.max())[0]
        if torch.sum(spikes) == 0:
            action[0] = torch.randint(low=0, high=1, size=(1,))[0]

        return action[0]

Action.register('highest_spikes_action', Highest_Spikes_Action)

class Highest_Voltage_Action(Action):

    """
    Selects an action that has the highest voltage. In case of equal spiking select randomly
    """
    steps_done = 0
    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Highest_Voltage_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.seed = kwargs.get('seed', 1)

    def torch_coding(self, record, target, device):
        """
        Args:
            record: voltage of output layer.
        Returns:
            Action sampled from highest voltage of output layer.
        """
        EPS_START = 0.9
        EPS_END = 0.05
        EPS_DECAY = 200

        assert (
           record.shape[2] == self.num
        ), "Output layer size is not equal to the size of the action space."

        import random
        import math
        random.seed(self.seed)
        sample = random.random()

        eps_threshold = EPS_END + (EPS_START - EPS_END) * math.exp(-1. * Highest_Voltage_Action.steps_done / EPS_DECAY)
        if record.shape[1] == 1:
            Highest_Voltage_Action.steps_done += 1
        if sample > eps_threshold:
            final_step_voltage = record[-1, :, :]
            action = final_step_voltage.max(1)[1]  # max(dim)返回的是value, index
            return action[0]
        else:
            return torch.randint(low=0, high=self.num, size=(1,))[0]

Action.register('highest_voltage_action', Highest_Voltage_Action)


class First_Spike_Action(Action):

    """
    Selects an action with the highst spikes. In case of equal spiking select randomly
    """

    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(First_Spike_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        """
        Args:
            record: spiking activity of output layer.
        Returns:
            Action sampled from first spike of output layer.
        """

        assert (
            record.shape[2] == self.num
        ), "Output layer size is not equal to the size of the action space."

        spikes = record.squeeze().nonzero()
        if spikes.shape[0] == 0:
            action = torch.randint(low=0, high=1, size=(1,))[0]
        else:
            action = spikes[0, 1]

        return action

Action.register('first_spike_action', First_Spike_Action)


class Random_Action(Action):

    """
    Selects an action randomly from the action space.
    """

    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Random_Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        """
        Used for the PyTorch backend

        Args:
           record: spiking activity of output layer.
        Returns:
           Action sampled from action space randomly.
       """
        return torch.randint(low=0, high=self.num, size=(1,))[0]

    def numpy_action(self, record):
        """
        Args:
           record: spiking activity of output layer.
        Returns:
           Action sampled from action space randomly.
       """
        return np.random.choice(self.num)

Action.register('random_action', Random_Action)



# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Generators.py
@time:2021/6/21 16:35
@description:
"""
from .Node import Node, Generator
import torch
import numpy as np

class Poisson_Generator(Generator):
    """
        泊松生成器，根据输入脉冲速率生成。
        Generate a poisson spike train according input rate.
        time: encoding window ms
        dt: time step
        HZ: cycles/s
    """
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), rate=None, **kwargs):

        super(Poisson_Generator, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.num = num
        # the unit of dt is 0.1ms, for each time step, the rate has to be multiplied by 1e-4
        self.unit_conversion = kwargs.get('unit_conversion', 0.1)
        self.weight = kwargs.get('weight', 1.0)
        self.start_time = kwargs.get('start_time', None)
        self.end_time = kwargs.get('end_time', None)
        if rate is not None:
            if hasattr(rate, '__iter__'):
                self.source = rate
            else:
                self.source = np.array([rate])
            self.new_input = True

        self.batch = kwargs.get('batch', 1)

    def torch_coding(self, source, device):

        if source.size != self.num:
            import warnings
            warnings.warn("The dimension of input data should be consistent with the number of input neurons.")
        if source.__class__.__name__ == 'ndarray':
            source = torch.tensor(source, device=device)
        self.inp_source = source
        #
        # # if source.ndim == 0:
        # #     batch = 1
        # # else:
        # #     batch = source.shape[SAS0]
        #
        # # shape = [self.batch, self.nu]
        # spk_shape = [self.time_step] + list(self.shape)
        # spikes = self.weight*torch.rand(spk_shape, device=device).le(source*self.unit_conversion).float()
        # if self.start_time is not None:
        #     start_time_step = int(self.start_time/self.dt)
        #     spikes[:start_time_step, ...] = 0.0
        # if self.end_time is not None:
        #     end_time_step = int(self.end_time/self.dt)
        #     spikes[end_time_step:, ...] = 0.0
        return None

    def next_stage(self):
        if self.new_input:
            self.get_input()
            self.shape[0] = self.inp_source.shape[0]
            self.new_input = False

        if (self.start_time is None or self.start_time< self.index*self.dt) \
            and (self.end_time is None or self.end_time> self.index*self.dt):
            spikes = self.weight*torch.rand(self.shape, device=self._backend.device[0]).le(
                self.inp_source*self.unit_conversion)
            return spikes.type(self._backend.data_type)
        else:
            return torch.zeros(self.shape, dtype=self._backend.data_type, device=self._backend.device)

Generator.register('poisson_generator', Poisson_Generator)

class Poisson_Generator2(Generator):
    """
        泊松生成器，根据输入脉冲速率生成。
        Generate a poisson spike train according input rate.
        time: encoding window ms
        dt: time step
    """
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):

        super(Poisson_Generator2, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.num = num




    def torch_coding(self, source, device):
        # assert (source >= 0).all(), "Input rate must be non-negative"
        if not (source >= 0).all():
            import warnings
            warnings.warn('Input rate shall be non-negative')
        if source.__class__.__name__ == 'ndarray':
            source = torch.tensor(source, device=device)

        if source.ndim == 0:
            batch = 1
        else:
            batch = source.shape[0]

        shape = list(self.shape)
        shape[0] = batch
        spk_shape = [self.time_step] + list(shape)
        spikes = torch.rand(spk_shape, device=device).le(source * self.dt).float()
        times = torch.zeros_like(spikes)
        spikes = torch.stack([spikes, times],dim=2)
        return spikes.type(self._backend.data_type)

Generator.register('poisson_generator2', Poisson_Generator2)


class CC_Generator(Generator):
    """
        恒定电流生成器。
        Generate a constant current input.
        time: encoding window ms
        dt: time step
    """
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None,
                 coding_method=('poisson_generator', 'cc_generator', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(CC_Generator, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, source, device):
        # assert (source >= 0).all(), "Input rate must be non-negative"
        if not (source >= 0).all():
            import warnings
            warnings.warn('Input rate shall be non-negative')
        if source.__class__.__name__ == 'ndarray':
            source = torch.tensor(source, dtype=self._backend.data_type, device=device)

        spk_shape = [self.time_step] + list(self.shape)
        spikes = source * torch.ones(spk_shape, device=device)
        return spikes.type(self._backend.data_type)

Generator.register('cc_generator', CC_Generator)
Generator.register('constant_current', CC_Generator)
class Sin_Generator(Generator):
    """

        Generate a sin current input.
        time: encoding window ms
        dt: time step
    """
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None,
                 coding_method=('poisson_generator', 'cc_generator', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Sin_Generator, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.num = num

    def torch_coding(self, source, device):
        # assert (source >= 0).all(), "Input rate must be non-negative"
        if source.__class__.__name__ == 'ndarray':
            source = torch.tensor(source, dtype=torch.float, device=device)
        amp = source[0]
        omg = 2*np.pi/source[1]
        # if source.ndim == 0:
        #     batch = 1
        # else:
        #     batch = source.shape[0]
        #
        # shape = [batch, self.num]
        spk_shape = [self.time_step] + [1 for _ in range(len(list(self.shape)))]

        t = torch.arange(0, self.time_step*self.dt, self.dt, device=device).view(spk_shape)
        spikes = amp*torch.sin(omg*t)
        return spikes

Generator.register('sin_generator', Sin_Generator)
Generator.register('sin', Sin_Generator)

class Ramp_Generator(Generator):
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None,
                 coding_method=('poisson_generator', 'cc_generator', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Ramp_Generator, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.base = kwargs.get('base', 0.0)
        self.end_time = kwargs.get('end_time', None)
        self.amp = kwargs.get('amp', 0.001)

    def torch_coding(self, source, device):
        if source.__class__.__name__ == 'ndarray':
            source = torch.tensor(source, dtype=torch.float, device=device)
        if self.base.__class__.__name__ == 'ndarray':
            self.base = torch.tensor(self.base, dtype=torch.float, device=device)

        slope = source
        # spk_shape = [self.time_step] + [1 for _ in range(len(list(self.shape)))]
        t_shape = [self.time_step] + [1 for _ in range(len(list(self.shape)))]
        t = torch.arange(0, self.time_step*self.dt, self.dt, device=device).view(t_shape)
        spikes = self.amp*(slope*t + self.base)
        if self.end_time is not None:
            time_step = int(self.end_time/self.dt)
            spikes[time_step:,...] = 0.0
        return spikes
Generator.register('ramp_generator', Ramp_Generator)
Generator.register('ramp', Ramp_Generator)









import os
import psutil
import sys
import torch

def get_cpu_mem():
    '''
    Get current process memory size in MiB.
    '''
    memory = round(psutil.Process(os.getpid()).memory_info().rss / 1048576, 4)
    return memory

def get_tens_mem(tensor):
    '''
    Get pytorch tensor memory size in MiB. 
    '''
    if torch.is_tensor(tensor):
        a = tensor.element_size() * tensor.nelement()
    else:
        a = 0
    tensor_memory = round(a / 1048576, 4)
    return tensor_memory

def get_object_size(object):
    '''
    Get object size in MiB.
    '''
    size = round(sys.getsizeof(object) / 1048576, 4)
    return size


# -*- coding: utf-8 -*-
"""
Created on 2020/8/11
@project: SPAIC
@filename: Node
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
定义神经网络的输入输出接口
"""
from spaic.Network.Assembly import Assembly
import torch
import numpy as np
from spaic.Backend.Backend import Op
class Node(Assembly):
    '''Base class for input encoder and output decoders.
    '''
    _class_label = '<nod>'
    _is_terminal = True

    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Node, self).__init__()

        self._dt = dt
        self._time = kwargs.get('time', None)
        self.coding_var_name = coding_var_name
        position = kwargs.get('position', None)
        if position == None:
            self.position = []
        else:
            position = np.array(position)
            assert position.shape[0] == num, " Neuron_position not equal to neuron number"
            self.position = position

        if coding_method == ('poisson', 'spike_counts', '...'):
            raise ValueError('Please specify the coding method such as poisson or spike_counts')
        else:
            self.coding_method = coding_method.lower()  # The name of coding method

        self.coding_var_name = coding_var_name

        if coding_method.lower() == 'null':
            self.is_encoded = True
        else:
            self.is_encoded = kwargs.get('is_encoded', False)

        # 单神经元多脉冲的语音数据集的shape包含脉冲时间以及发放神经元标签，所以不能通过np.prod(shape)获取num，最好还是外部输入num
        assert num is not None or shape is not None, "One of the shape and number must not be None"
        if num is None:
            if coding_method == 'mstb' or coding_method == 'sstb':
                raise ValueError('Please set the number of node')
            self.num = np.prod(shape)
        else:
            self.num = num

        self.num = int(self.num)  # 统一数据格式为Python内置格式

        if shape is None:
            if self.is_encoded:
                self.shape = [1, 1, self.num]
            else:
                self.shape = [1, self.num]
        else:
            if coding_method == 'mstb' or coding_method == 'sstb':
                self.shape = (1, self.num)
            elif self.is_encoded:
                self.shape = [1, 1] + list(shape)
            else:
                self.shape = [1] + list(shape)

        if node_type == ('excitatory', 'inhibitory', 'pyramidal', '...'):
            self.type = []
        elif isinstance(node_type, list):
            self.type = node_type
        else:
            self.type = [node_type]



        # Coding parameters
        self.coding_param = kwargs
        self.dec_target = dec_target
        self.source = np.random.rand(*self.shape)
        #TODO:现在 Encoder, Decoder, Generator功能拆开了，是不是应该把这段放到Decoder类里了, 上边的coding_method也是同理
        if self.dec_target is not None:
            # The size of predict, reward and action is equal to batch_size
            self.predict = np.zeros((1,))
            # self.reward = np.zeros((1,))
            # self.action = np.zeros((1,))

        # Parameters of initial operation
        self.index = 0
        self.records = []
        self._var_names = list()

    def init_state(self):
        self.index = 0

    @property
    def dt(self):
        if self._dt is None and self._backend is not None:
            return self._backend.dt
        else:
            return self._dt

    @dt.setter
    def dt(self, dt):
        self._dt = dt

    @property
    def time(self):
        if self._time is None and self._backend is not None:
            return self._backend.runtime
        else:
            return self._time

    @property
    def time_step(self):
        return int(np.ceil(self.time / self.dt))


    @time.setter
    def time(self, time):
        self._time = time

    def get_var_names(self):
        return self._var_names

    @staticmethod
    def register(name, coding_class):
        '''
        Register a coding class. Registered encoding or decoding classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'poisson', 'spike_counts')
        coding_class :
            The subclass of coding object, e.g. an 'PoissonEncoding', 'Spike_Counts'.
        '''

        raise NotImplementedError

    def torch_coding(self, source: torch.Tensor, target: torch.Tensor , device: str) -> torch.Tensor:
        '''
        Args:
            source : It is input spike trains for encoding class and output spike trains for decoding class.
            target : It is None  for encodoing class and labels for decoding class.
            device : CPU or CUDA, this parameter is taken from backend
        Returns:
        '''

        raise NotImplementedError

    def numpy_coding(self, source, target, device):
        raise NotImplementedError

    def tensorflow_coding(self, source, target, device):
        raise NotImplementedError

    def build(self, backend):
        self._backend = backend
        self.data_type = backend.data_type

    def __call__(self, data=None):

        if isinstance(self, Encoder) or isinstance(self, Generator) or isinstance(self, Reward):
            if isinstance(data, np.ndarray):
                self.source = data
                batch_size = data.shape[0]
            elif isinstance(data, torch.Tensor):
                self.source = data
                batch_size = data.shape[0]
            elif isinstance(data, list) and self.coding_method=="mstb":
                self.source = data
                batch_size = len(self.source)
            elif hasattr(data, '__iter__'):
                self.source = np.array(data)
                batch_size = self.source.shape[0]
            else:
                self.source = np.array([data])
                batch_size = 1

            if self._backend is None:
                self.batch_size = batch_size
            else:
                self._backend.set_batch_size(batch_size)
                self.batch_size = None

            self.new_input = True

        elif isinstance(self, Decoder):
            if isinstance(data, np.ndarray):
                self.source = data
            elif isinstance(data, torch.Tensor):
                self.source = data
            elif hasattr(data, '__iter__'):
                self.source = np.array(data)
            else:
                self.source = np.array([data])

            return self.predict


# ======================================================================================================================
# Encoders
# ======================================================================================================================
class Encoder(Node):
    '''
        Five encoding method are provided, as shown below (key: class):
        1. 'sstb': SigleSpikeToBinary,
        2. 'mstb': MultipleSpikeToBinary
        3. 'poisson': PoissonEncoding
        4. 'latency': Latency
        5. 'relative_latency': Relative_Latency
    '''
    _node_sub_class = '<encoder>'
    _coding_subclasses = dict()
    def __init__(self, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Encoder, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.batch_size = None
        self.new_input = True
        # coding_method = coding_method.lower()
        # if coding_method == 'null':
        #     self.is_encoded = True
        # else:
        #     self.is_encoded = kwargs.get('is_encoded', False)

    def __new__(cls, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        coding_method = coding_method.lower()

        if cls is not Encoder:
            return super().__new__(cls)
        elif coding_method in Encoder._coding_subclasses:
            return super().__new__(Encoder._coding_subclasses[coding_method])
        else:
            raise ValueError("No coding method: %s in Encoding classes" % coding_method)

    @staticmethod
    def register(name, coding_class):
        '''
        Register a coding class. Registered encoding or decoding classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'poisson', 'spike_counts')
        coding_class :
            The subclass of coding object, e.g. an 'PoissonEncoding', 'Spike_Counts'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Encoder._coding_subclasses:
            raise ValueError(('A coding class with the name "%s" has already been registered') % name)

        if not issubclass(coding_class, Encoder):
            raise ValueError(
                ('Given class of type %s does not seem to be a valid encoding class.' % str(type(coding_class))))

        Encoder._coding_subclasses[name] = coding_class

    def init_state(self):
        self.index = 0

    # initial operation: encoding input features into spike patterns
    def get_input(self):
        self.index = 0
        if self.sim_name == 'pytorch':
            spikes = self.torch_coding(self.source, self.device)
        else:
            spikes = self.numpy_coding(self.source, self.device)
        self.all_spikes = spikes
        return self.all_spikes

    # stand alone operation: get spike pattern in every time step
    def next_stage(self):
        # For hardware applications, call next_stage at each time step to get spike data of the current time step.
        if self.new_input:
            self.get_input()
            self.new_input = False

        self.index += 1

        return self.all_spikes[self.index-1]

    def reset(self):
        # Called at the start of each epoch
        self.init_state()

    def build(self, backend):
        self._backend = backend
        self.sim_name = backend.backend_name
        self.device = backend.device0

        # if self.dt is None:
        #     self.dt = backend.dt
        if self.batch_size is not None:
            self._backend.set_batch_size(self.batch_size)

        # if self.sim_name == 'pytorch':
        #     spikes = self.torch_coding(self.source, self.device)  # (time_step, batch_size, shape)
        # else:
        #     spikes = self.numpy_coding(self.source, self.device)
        # self.all_spikes = spikes

        if self.is_encoded:
            shape = self.shape[1:]
        else:
            shape = self.shape
        # self.shape = spikes[0].shape

        key = self.id + ':' + '{'+self.coding_var_name+'}'
        self.variable_to_backend(key, shape, value=0)
        self.init_op_to_backend(None, self.init_state, [])
        backend.register_standalone(Op(key, self.next_stage, [], owner=self))


# ======================================================================================================================
# Decoders
# ======================================================================================================================

class Decoder(Node):
    '''
        Five decoding method are provided, as shown below (key: class):
        1. 'spike_counts': Spike_Counts
        2. 'first_spike': First_Spike
        3. 'time_spike_counts': TimeSpike_Counts
        4. 'time_softmax': Time_Softmax
        5. 'final_step_voltage': Final_Step_Voltage
    '''
    _node_sub_class = '<decoder>'
    _coding_subclasses = dict()
    def __init__(self, num=None, dec_target=None,  dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Decoder, self).__init__(None, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        assert num == dec_target.num, ('The num of Decoder is not consistent with num of NeuronGroup')

    def __new__(cls, num=None, dec_target=None, dt=None, coding_method='spike_counts',
                coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        coding_method = coding_method.lower()
        if cls is not Decoder:
            return super().__new__(cls)
        elif coding_method in Decoder._coding_subclasses:
            return super().__new__(Decoder._coding_subclasses[coding_method])
        else:
            raise ValueError("No coding method: %s in Decoding classes" % coding_method)

    @staticmethod
    def register(name, coding_class):
        '''
        Register a coding class. Registered encoding or decoding classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'poisson', 'spike_counts')
        coding_class :
            The subclass of coding object, e.g. an 'PoissonEncoding', 'Spike_Counts'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Decoder._coding_subclasses:
            raise ValueError(('A coding class with the name "%s" has already been registered') % name)

        if not issubclass(coding_class, Decoder):
            raise ValueError(
                ('Given class of type %s does not seem to be a valid decoding class.' % str(type(coding_class))))

        Decoder._coding_subclasses[name] = coding_class

    def init_state(self):
        self.index = 0

    # stand alone operation: decoding spike patterns. Predict can be predict labels or RL action
    def get_output(self, output):
        if (self.index % self.time_step) == 0:
            shape = list(output.shape)
            dec_shape = [self.time_step] + shape
            if type(output).__name__ == 'Tensor':
                self.records = torch.zeros(dec_shape, device=self.device, dtype=output.dtype)
            else:
                self.records = np.zeros(dec_shape)
            self.index = 0

        self.records[self.index % self.time_step, :] = output
        self.index += 1
        if self.index >= self.time_step:
            if self.sim_name == 'pytorch':
                self.predict = self.torch_coding(self.records, self.source, self.device)
            else:
                self.predict = self.numpy_coding(self.records, self.source, self.device)
        return 0

    def reset(self):
        # Called at the start of each epoch
        self.init_state()

    def build(self, backend):
        self._backend = backend
        self.sim_name = backend.backend_name
        self.device = backend.device0
        # if self.dt is None:
        #     self.dt = backend.dt

        output_name = self.dec_target.id + ':' + '{'+self.coding_var_name+'}'
        self.init_op_to_backend(None, self.init_state, [])
        backend.register_standalone(Op(None, self.get_output, [output_name], owner=self))

# ======================================================================================================================
# Rewards
# ======================================================================================================================

class Reward(Node):
    '''
        Three reward method are provided, as shown below (key: class):
        1. 'global_reward', Global_Reward
        2. 'xor_reward': XOR_Reward
        3. 'da_reward': DA_Reward
        4. 'environment_reward': Environment_Reward
    '''
    _node_sub_class = '<reward>'
    _coding_subclasses = dict()
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Reward, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.dec_sample_step = kwargs.get('dec_sample_step', 1)
        self.reward_shape = kwargs.get('reward_shape', (1, ))

    def __new__(cls, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        coding_method = coding_method.lower()
        if cls is not Reward:
            return super().__new__(cls)
        elif coding_method in Reward._coding_subclasses:
            return super().__new__(Reward._coding_subclasses[coding_method])
        else:
            raise ValueError("No reward method: %s in Reward classes" % coding_method)

    @staticmethod
    def register(name, coding_class):
        '''
        Register a Reward class. Registered reward classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'step_reward')
        coding_class :
            The subclass of coding object, e.g. an 'Step_Reward'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Reward._coding_subclasses:
            raise ValueError(('A reward class with the name "%s" has already been registered') % name)

        if not issubclass(coding_class, Reward):
            raise ValueError(
                ('Given class of type %s does not seem to be a valid reward class.' % str(type(coding_class))))

        Reward._coding_subclasses[name] = coding_class

    def init_state(self):
        self.index = 0

    # stand alone operation: get reward from output activities
    def get_reward(self, output=np.empty(0)):
        self.device = self._backend.device0
        if (self.index % self.dec_sample_step) == 0:
            self.index = 0
            shape = list(output.shape)
            dec_shape = [self.dec_sample_step] + shape
            if type(output).__name__ == 'Tensor':
                self.records = torch.zeros(dec_shape, device=self.device)
            else:
                self.records = np.zeros(dec_shape)
        reward = torch.zeros(self.reward_shape, device=self.device)
        self.records[self.index, :] = output
        self.index += 1
        if self.index >= self.dec_sample_step:
            if self.sim_name == 'pytorch':
                reward = self.torch_coding(self.records, self.source, self.device)
                self.reward = reward
            else:
                reward = self.numpy_coding(self.records, self.source, self.device)
        return reward

    def build(self, backend):
        self._backend = backend
        self.sim_name = backend.backend_name
        self.data_type = backend.data_type
        self.device = backend.device0
        # if self.dt is None:
        #     self.dt = backend.dt
        self.init_op_to_backend(None, self.init_state, [])
        reward_name = 'Output_Reward'
        self.variable_to_backend(reward_name, self.reward_shape, value=0.0)  # shape还是要让具体的子类定义吧
        if self.dec_target is not None:
            output_name = self.dec_target.id + ':' + '{'+self.coding_var_name+'}'
            backend.register_standalone(Op(reward_name, self.get_reward, [output_name], owner=self))
        else:
            backend.register_standalone(Op(reward_name, self.get_reward, [], owner=self))


class Generator(Node):
    '''
        Two generator method are provided, as shown below (key: class):
        1. 'poisson_generator': Poisson_Generator,
        2. 'cosine_generator': Cosine_Generator
    '''
    _node_sub_class = '<generator>'
    _coding_subclasses = dict()
    def  __init__(self, shape=None, num=None, dec_target=None,  dt=None,
                  coding_method=('poisson_generator', 'cc_generator', '...'),
                  coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Generator, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.gen_first = kwargs.get("gen_first", False) #only generate data for the first time, then use this date for later usage
        self.all_spikes = None
        self.build_level = 0
        self.new_input = True

    def __new__(cls, shape=None, num=None, dec_target=None, dt=None,
                coding_method=('poisson_generator', 'cc_generator', '...'),
                coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        if cls is not Generator:
            return super().__new__(cls)
        elif coding_method in Generator._coding_subclasses:
            return super().__new__(Generator._coding_subclasses[coding_method])
        else:
            raise ValueError("No coding method: %s in Generator classes" % coding_method)

    @staticmethod
    def register(name, coding_class):
        '''
        Register a coding class. Registered encoding or decoding classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'poisson', 'spike_counts')
        coding_class :
            The subclass of coding object, e.g. an 'PoissonEncoding', 'Spike_Counts'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Generator._coding_subclasses:
            raise ValueError(('A coding class with the name "%s" has already been registered') % name)

        if not issubclass(coding_class, Generator):
            raise ValueError(
                ('Given class of type %s does not seem to be a valid generator class.' % str(type(coding_class))))

        Generator._coding_subclasses[name] = coding_class

    def init_state(self):
        self.index = 0
        self.new_input = True

    def torch_coding(self, source, device):
        '''

        Args:
            source (): It is input spike trains for encoding class and output spike trains for decoding class.
            device (): CPU or CUDA, this parameter is taken from backend

        Returns:

        '''
        raise NotImplementedError

    # initial operation: encoding input features into spike patterns
    def get_input(self):
        self.index = 0
        if (self.gen_first is True and self.all_spikes is None) or self.gen_first is False:
            if self.sim_name == 'pytorch':
                spikes = self.torch_coding(self.source, self.device)
            else:
                spikes = self.numpy_coding(self.source, self.device)
            self.all_spikes = spikes
        return self.all_spikes

    # stand alone operation: get spike pattern in every time step
    def next_stage(self):
        if self.new_input:
            self.get_input()
            self.new_input = False
        self.index += 1
        return self.all_spikes[self.index-1]

    def build(self, backend):
        self._backend = backend
        self.sim_name = backend.backend_name
        self.device = backend.device0
        # if self.dt is None:
        #     self.dt = backend.dt

        if self.sim_name == 'pytorch':
            singlestep_spikes = torch.zeros(self.shape, device=self.device)
        else:
            singlestep_spikes = np.zeros(self.shape)

        if self.dec_target is None:
            key = self.id + ':' + '{'+self.coding_var_name+'}'
            self.variable_to_backend(key, self.shape, value=singlestep_spikes)
        else:
            key = self.dec_target.id + ':' + '{'+self.coding_var_name+'}'
        self._var_names.append(key)
        self.init_op_to_backend(None, self.init_state, [])
        backend.register_standalone(Op(key, self.next_stage, [], owner=self))


# ======================================================================================================================
# Actions
# ======================================================================================================================

class Action(Node):
    '''
        Six action method are provided, as shown below (key: class):
        1. 'pop_rate_action': PopulationRate_Action
        2. 'softmax_action': Softmax_Action
        3. 'highest_spikes_action': Highest_Spikes_Action
        4. 'highest_voltage_action', Highest_Voltage_Action
        5. 'first_spike_action': First_Spike_Action
        6. 'random_action': Random_Action
    '''
    _node_sub_class = '<action>'
    _coding_subclasses = dict()
    def __init__(self, shape=None, num=None, dec_target=None,  dt=None, coding_method=('poisson', 'spike_counts', '...'),
                 coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Action, self).__init__(shape, num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.action = np.zeros((1,))


    def __new__(cls, shape=None, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'),
                coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        coding_method = coding_method.lower()
        if cls is not Action:
            return super().__new__(cls)
        elif coding_method in Action._coding_subclasses:
            return super().__new__(Action._coding_subclasses[coding_method])
        else:
            raise ValueError("No action method: %s in Action classes" % coding_method)

    @staticmethod
    def register(name, coding_class):
        '''
        Register an action class. Registered action classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. 'pop_rate_action')
        coding_class :
            The subclass of coding object, e.g. an 'PopulationRate_Action'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Action._coding_subclasses:
            raise ValueError(('A action class with the name "%s" has already been registered') % name)

        if not issubclass(coding_class, Action):
            raise ValueError(
                ('Given class of type %s does not seem to be a valid action class.' % str(type(coding_class))))

        Action._coding_subclasses[name] = coding_class

    def init_state(self):
        self.index = 0

    # stand alone operation: decoding spike patterns. Output is RL action
    def get_action(self, output):
        if (self.index % self.time_step) == 0:
            shape = list(output.shape)
            dec_shape = [self.time_step] + shape
            if type(output).__name__ == 'Tensor':
                self.records = torch.zeros(dec_shape, device=self.device)
            else:
                self.records = np.zeros(dec_shape)
        self.records[self.index % self.time_step, :] = output
        self.index += 1
        if self.index >= self.time_step:
            if self.sim_name == 'pytorch':
                self.action = self.torch_coding(self.records, self.source, self.device)
            else:
                self.action = self.numpy_coding(self.records, self.source, self.device)
        return 0

    def build(self, backend):
        self._backend = backend
        self.sim_name = backend.backend_name
        self.device = backend.device0
        # if self.dt is None:
        #     self.dt = backend.dt

        output_name = self.dec_target.id + ':' + '{'+self.coding_var_name+'}'
        self.init_op_to_backend(None, self.init_state, [])
        backend.register_standalone(Op(None, self.get_action, [output_name], owner=self))



from .Learner import Learner
import torch
import numpy as np

class Tempotron(Learner):
    '''
        Tempotron learning rule.
        Args:
            tau(num) : The parameter tau of tempotron learning model.
            tau_s(num) : The parameter tau_s of tempotron learning model.
            V0(num) : The parameter V0 of tempotron learning model.
            dw_i(num) : The parameter dw_i of tempotron learning model.
            time_window(num) : The parameter time_window of tempotron learning model.
            preferred_backend(list) : The backend prefer to use, should be a list.
            name(str) : The name of this learning model. Should be 'Tempotron'.
        Methods:
            initial_param(self, weight): initialize some parameters for each batch.
            tempotron_update(self, input, output, weight, output_V): calculate the update of weight
            build(self, simulator): Build the simulator, realize the algorithm of Tempotron learning model.
        Example:
            self._learner = BaseLearner(algorithm='Tempotron', lr=0.5, trainable=self, conn=self.connection1)
        Reference:
            The tempotron: a neuron that learns spike timing_based decisions.
            doi: 10.1038/nn1643.
            url: http:www.nature.com/natureneuroscience
    '''

    def __init__(self, trainable=None, lr=0.01, **kwargs):
        super(Tempotron, self).__init__(trainable=trainable)

        self.prefered_backend = ['pytorch']
        self.firing_func = None
        self.lr = lr
        self.tau = 20
        self.tau_s = 5
        self.required_grad = kwargs.get('required_grad')
        self.trainable = trainable

        self.fake = 1
        self.v_th = 10
        self.V_max = 0

    def initial_param(self, weight):
        '''
            Initialize some parameters for each batch.
            Args:
                weight: the weight for the connection which needed to learned.
        '''
        self.total_V = []
        self.run_time = self._backend.runtime
        self.total_Input = []
        self.total_Output = []
        self.total_step = self.run_time / self.dt - 1
        self.dw = torch.zeros(weight.shape)
        dw_i = [np.exp(-t / self.tau) - np.exp(-t / self.tau_s) for t in range(int(self.run_time / self.dt))]
        # dw_i = dw_i[:: -1]  # 将dw翻转一下
        # self.V0 = 1 / (max(dw_i))
        self.V0 = 6.75 # V0
        # self.dw_i = torch.mul(self.V0.to(self.device), torch.tensor(dw_i).to(self.device))
        self.dw_i = torch.tensor(dw_i).to(self.device)

        return self.dw_i

    def tempotron_update(self, input, output, weight, output_V):
        '''
            Args:
                input: input spikes.
                output: output spikes.
                weight: the weight for the connection which needed to learned.
                output_V: the voltage of postsynaptic neurons
        Returns:
             Updated weight.
        '''

        if self.fake:
            self.fake -= 1
            return weight
        input = torch.unsqueeze(input, 0)
        output = torch.unsqueeze(output, 0)
        output_V = torch.unsqueeze(output_V, 0)
        output_spikes = output.data
        input_spikes = input.data
        # output_num = self.trainable.output.num
        input_num = self.trainable[0].pre.num

        if self.total_V == []:
            self.total_V = output_V
        else:
            self.total_V = torch.cat((self.total_V, output_V), dim=0)

        if self.total_Input == []:
            self.total_Input = input_spikes
        else:
            self.total_Input = torch.cat((self.total_Input, input_spikes), dim=0)
        if self.total_Output == []:
            self.total_Output = output_spikes
        else:
            self.total_Output = torch.cat((self.total_Output, output_spikes), dim=0)

        if self._backend.n_time_step < self.total_step:
            pass

        else:

            label = self.trainable[1].source
            b = torch.zeros(weight.shape)

            # output_label = (torch.argmax(torch.sum(self.total_Output, dim=0), dim=1)) #先将所有时间步的output求和，然后再根据output那一列找到最大的output索引
            # self.total_output维度(总时间步,batch_size, output_num)

            update = []

            for i in range(output_V.shape[1]):  # 把每个batch分开算
                Input_each = self.total_Input[:, i, :]
                output_each = self.total_Output[:, i, :]
                V_each = self.total_V[:, i, :]
                # import matplotlib.pyplot as plt
                # plt.plot(V_each[:, 0].cpu().detach().numpy())
                #
                # plt.show()

                # 判断output的tmax，看v的最大值和output发放的时间点谁更早
                # t_max_all = torch.argmax(output_each, dim=0) if 1 in output else torch.argmax(V_each, dim=0)
                t_max_all = torch.argmax(V_each, dim=0)  # 十个output神经元的最大V的index

                # 最大值的V有没有超过阈值，如果超过阈值，那么不动，如果没超过阈值，那么V要增加
                # 除了最大值的V其他的V都不应该超过阈值，如果有超过阈值的，需要减小V
                # t_o = torch.where(output_each)

                for o in range(t_max_all.shape[0]):

                    V = V_each[t_max_all[o], o]
                    # print(V)
                    if o == label[i] and V < self.v_th :

                        b[o, :] = 1

                    elif o != label[i] and V >= self.v_th :

                        b[o, :] = -1
                # Input_dw = torch.zeros_like(weight)
                for o in range(t_max_all.shape[0]):
                    a_1 = torch.linspace(0, t_max_all[o] - 1, steps=t_max_all[o]).to(
                        self.device).type(torch.long)
                    Input_window = torch.index_select(Input_each, 0, a_1).permute(1, 0)  # (784, t_max_all[i][j])
                    dw = torch.flip(self.dw_i[:t_max_all[o]], dims=[0])
                    # dw = self.dw_i[:t_max_all[o]][::-1]
                    Input_dw_each = torch.sum((Input_window * dw), dim=1)
                    self.dw[o] = Input_dw_each
                    # if t_max_all[o] - self.time_window + 1 > 0:
                    #     a = torch.linspace(t_max_all[o] - self.time_window + 1, t_max_all[o],
                    #                        steps=self.time_window).to(
                    #         self.device).type(torch.long)
                    #     Input_window = torch.index_select(Input_each, 0, a).permute(1, 0)  # (784, self.t)
                    # else:
                    #     a_1 = torch.linspace(0, t_max_all[o], steps=t_max_all[o]+1).to(
                    #         self.device).type(torch.long)
                    #     Input_window = torch.index_select(Input_each, 0, a_1).permute(1, 0)  # (784, t_max_all[i][j])
                    #     dw = torch.flip(self.dw_i[:t_max_all[o] +1], dims=[0])
                    #     # dw = self.dw_i[:t_max_all[o]][::-1]
                    #     Input_dw_each = torch.sum((Input_window * dw), dim=1)
                    #     self.dw[o] = Input_dw_each
                    #     # pad_num = self.time_window - t_max_all[o] - 1
                    #     # pad = torch.zeros((input_num, pad_num)).to(self.device)
                    #     # Input_window = torch.cat((Input_window, pad), 1)

                    # c = Input_window * self.dw_i.to(self.device)
                    # Input_dw = torch.sum(c, 1)
                    # Input_dw = torch.sum(Input_window * self.dw_i.to(self.device), 1)#(把每一个神经元时间窗里的值都加起来)
                    # self.dw[o, :] = Input_dw

                self.dw = self.lr * b * self.dw

                if update == []:
                    update = torch.unsqueeze(self.dw, 0)
                else:
                    update = torch.cat((update, torch.unsqueeze(self.dw, 0)), dim=0)
            update = torch.mean(update, dim=0)

            if self.required_grad == False:
                with torch.no_grad():
                    # print('update1', update)
                    weight.add_(update.to(self.device))
                    weight.clamp_(0.0, 2.0)
            else:
                # print('update2', update)
                weight[...] = weight + update.to(self.device)  # 注意tensor会出现假赋值的情况，就是他没有赋值给原来的tensor，而是自己新建了一个tensor
            # self.V_max = torch.max(self.total_V, dim=0).values
            # self._simulator._variables['autoname1<net>_layer1<neg>:{V_max[stay]}'] = self.V_max
            # print([self.V_max])
            # print('a')

        return weight

    def build(self, backend):
        '''
            Build the simulator, realize the algorithm of Tempotron model.
            Args:
                simulator: The simulator we used to compute.
        '''

        super(Tempotron, self).build(backend)
        self.device = backend.device0
        self._backend = backend
        self.sim_name = backend.backend_name
        self.dt = backend.dt

        if backend.backend_name == 'pytorch':
            import torch

            # 为online_stdp函数的input, output, weight 等参数命名（取名函数在Connection中）
            for conn in self.trainable_connections.values():
                preg = conn.pre
                postg = conn.post
                # pre_name = conn.get_input_name(preg, postg)
                # post_name = conn.get_target_output_name(postg)
                pre_name = conn.get_input_name(preg, postg)
                post_name = conn.get_group_name(postg, 'O')
                weight_name = conn.get_link_name(preg, postg, 'weight')

                V_name = conn.get_group_name(postg, 'V[updated]')
                # V0 = conn.get_group_name(postg, 'V0')
                # V_max = conn.get_group_name(postg, 'V_max')
                # V_name = conn.get_V_updated_name(postg)
                # V0 = conn.post_assembly.id + ':' + '{V0}'
                # V_max = conn.post_assembly.id + ':' + '{V_max}'

                # register_standalone(self,output_name: str, function, input_names(后端输入的名字): list):
                self.init_op_to_backend(None, self.initial_param, [weight_name])
                # simulator.add_variable(V_max, [], self.V_max)
                self.op_to_backend(weight_name, self.tempotron_update,
                                              [pre_name, post_name, weight_name, V_name])


Learner.register("Tempotron", Tempotron)


# Copyright 2014 Jason Heeris, jason.heeris@gmail.com
# 
# This file is part of the gammatone toolkit, and is licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING
"""
This module contains functions for calculating weights to approximate a
gammatone filterbank-like "spectrogram" from a Fourier transform.
"""
from __future__ import division
import numpy as np

from .filters import erb_space, make_erb_filters
from .gtgram import gtgram

def specgram_window(
        nfft,
        nwin,
    ):
    """
    Window calculation used in specgram replacement function. Hann window of
    width `nwin` centred in an array of width `nfft`.
    """
    halflen = nwin // 2
    halff = nfft // 2 # midpoint of win
    acthalflen = int(np.floor(min(halff, halflen)))
    halfwin = 0.5 * ( 1 + np.cos(np.pi * np.arange(0, halflen+1)/halflen))
    win = np.zeros((nfft,))
    win[halff:halff+acthalflen] = halfwin[0:acthalflen]
    win[halff:halff-acthalflen:-1] = halfwin[0:acthalflen]
    return win


def specgram(x, n, sr, w, h):
    """ Substitute for Matlab's specgram, calculates a simple spectrogram.
    :param x: The signal to analyse
    :param n: The FFT length
    :param sr: The sampling rate
    :param w: The window length (see :func:`specgram_window`)
    :param h: The hop size (must be greater than zero)
    """
    # Based on Dan Ellis' myspecgram.m,v 1.1 2002/08/04
    assert h > 0, "Must have a hop size greater than 0"

    s = x.shape[0]
    win = specgram_window(n, w)

    c = 0

    # pre-allocate output array
    ncols = 1 + int(np.floor((s - n)/h))
    d = np.zeros(((1 + n // 2), ncols), np.dtype(complex))

    for b in range(0, s - n, h):
      u = win * x[b : b + n]
      t = np.fft.fft(u)
      d[:, c] = t[0 : (1 + n // 2)].T
      c = c + 1

    return d


def fft_weights(
    nfft,
    fs,
    nfilts,
    width,
    fmin,
    fmax,
    maxlen):
    """
    :param nfft: the source FFT size
    :param sr: sampling rate (Hz)
    :param nfilts: the number of output bands required (default 64)
    :param width: the constant width of each band in Bark (default 1)
    :param fmin: lower limit of frequencies (Hz)
    :param fmax: upper limit of frequencies (Hz)
    :param maxlen: number of bins to truncate the rows to
    
    :return: a tuple `weights`, `gain` with the calculated weight matrices and
             gain vectors
    
    Generate a matrix of weights to combine FFT bins into Gammatone bins.
    
    Note about `maxlen` parameter: While wts has nfft columns, the second half
    are all zero. Hence, aud spectrum is::
    
        fft2gammatonemx(nfft,sr)*abs(fft(xincols,nfft))
    
    `maxlen` truncates the rows to this many bins.
    
    | (c) 2004-2009 Dan Ellis dpwe@ee.columbia.edu  based on rastamat/audspec.m
    | (c) 2012 Jason Heeris (Python implementation)
    """
    ucirc = np.exp(1j * 2 * np.pi * np.arange(0, nfft / 2 + 1) / nfft)[None, ...]
    
    # Common ERB filter code factored out
    cf_array = erb_space(fmin, fmax, nfilts)[::-1]

    _, A11, A12, A13, A14, _, _, _, B2, gain = (
        make_erb_filters(fs, cf_array, width).T
    )
    
    A11, A12, A13, A14 = A11[..., None], A12[..., None], A13[..., None], A14[..., None]

    r = np.sqrt(B2)
    theta = 2 * np.pi * cf_array / fs    
    pole = (r * np.exp(1j * theta))[..., None]
    
    GTord = 4
    
    weights = np.zeros((nfilts, nfft))

    weights[:, 0:ucirc.shape[1]] = (
          np.abs(ucirc + A11 * fs) * np.abs(ucirc + A12 * fs)
        * np.abs(ucirc + A13 * fs) * np.abs(ucirc + A14 * fs)
        * np.abs(fs * (pole - ucirc) * (pole.conj() - ucirc)) ** (-GTord)
        / gain[..., None]
    )

    weights = weights[:, 0:int(maxlen)]

    return weights, gain


def fft_gtgram(
    wave,
    fs,
    window_time, hop_time,
    channels,
    f_min):
    """
    Calculate a spectrogram-like time frequency magnitude array based on
    an FFT-based approximation to gammatone subband filters.
    A matrix of weightings is calculated (using :func:`gtgram.fft_weights`), and
    applied to the FFT of the input signal (``wave``, using sample rate ``fs``).
    The result is an approximation of full filtering using an ERB gammatone
    filterbank (as per :func:`gtgram.gtgram`).
    ``f_min`` determines the frequency cutoff for the corresponding gammatone
    filterbank. ``window_time`` and ``hop_time`` (both in seconds) are the size
    and overlap of the spectrogram columns.
    | 2009-02-23 Dan Ellis dpwe@ee.columbia.edu
    |
    | (c) 2013 Jason Heeris (Python implementation)
    """
    width = 1 # Was a parameter in the MATLAB code

    nfft = int(2 ** (np.ceil(np.log2(2 * window_time * fs))))
    nwin, nhop, _ = gtgram.gtgram_strides(fs, window_time, hop_time, 0)

    gt_weights, _ = fft_weights(
            nfft,
            fs,
            channels,
            width,
            f_min,
            fs / 2,
            nfft / 2 + 1
        )

    sgram = specgram(wave, nfft, fs, nwin, nhop)

    result = gt_weights.dot(np.abs(sgram)) / nfft

    return result


# -*- coding: utf-8 -*-
"""
Created on 2020/8/11
@project: SPAIC
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
# print('__file__={0:<35} | __name__={1:<20} | __package__={2:<20}'.format(__file__,__name__,str(__package__)))
from .Neuron import NeuronGroup, NeuronModel
from .Node import Node




# -*- coding: utf-8 -*-
"""
Created on 2022/1/18
@project: SPAIC
@filename: Topology
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""


from spaic.Network.BaseModule import BaseModule
from spaic.Network.Assembly import Assembly
# from spaic.Network.Synapse import Flatten_synapse
from spaic.IO.Initializer import BaseInitializer
from spaic.Network.BaseModule import VariableAgent, Op
from spaic.Backend.Backend import Backend
from abc import abstractmethod
from typing import List
import numpy as np
from abc import ABC
from spaic.utils.memory import get_cpu_mem,get_object_size

class Projection(BaseModule):
    '''
    Class for projection between assemblies, which contain multiple connections between sub neurongroups
    '''
    _class_label = '<prj>'
    _con_count = 0

    def __init__(self, pre: Assembly, post: Assembly, policies=[], link_type=None, ConnectionParameters=None,
                 name=None):
        super(Projection, self).__init__()
        assert isinstance(pre, Assembly)
        assert isinstance(post, Assembly)
        self.name = name
        self.super = None
        self.abstract_level = 0
        self._backend: Backend = None
        self.pre = pre
        self.post = post
        self._connections = dict()
        self._projections = dict()

        if (pre.id is not None) and (post.id is not None):
            self.key = pre.id + "->" + post.id
        else:
            self.key = 'default' + str(Projection._con_count)
        Projection._con_count += 1
        self._leaf_connections = dict()

        if isinstance(ConnectionParameters, dict):
            self.ConnectionParameters = ConnectionParameters
        else:
            self.ConnectionParameters = dict()
        if pre._is_terminal and post._is_terminal:
            self.is_unit = True
        else:
            self.is_unit = False

        self._policies = policies
        if self._policies:
            assert link_type is not None
        if isinstance(self._policies, ConnectPolicy):
            self._policies = [self._policies]
        self.link_type = link_type
        self._supers = []

    def __setattr__(self, name, value):
        super(Projection, self).__setattr__(name, value)
        if name == 'super':
            return
        if isinstance(value, Connection):
            if self._backend: self._backend.builded = False
            self._connections[name] = value
            # self._leaf_connections[value.key] = value
            value.set_name(name)
            value.add_super(self)
        elif isinstance(value, Projection):
            if self._backend: self._backend.builded = False
            self._projections[name] = value
            # self._leaf_connections[value.key] = value
            value.set_name(name)
            value.add_super(self)

    def homologous(self, other):
        if (other.pre is self.pre) and (other.post is self.post):
            return True
        else:
            return False

    def is_empty(self):
        if self.is_unit:
            return False
        elif len(self._connections) + len(self._projections) + len(self._leaf_connections) == 0:
            return True
        else:
            return False

    def __and__(self, other):
        if not self.homologous(other):
            raise ValueError("can't do & operation for nonhomologous connections")
        if self.is_empty():
            new_connect = self
        elif other.is_empty():
            new_connect = other
        else:
            new_connect = Projection(self.pre, self.post)
            # unit connections
            key1 = set(self._connections.keys())
            key2 = set(other._connections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._connections[key] = self._connections[key]
                new_connect._connections[key].super = new_connect

            # assb connections
            key1 = set(self._projections.keys())
            key2 = set(other._projections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._projections[key] = self._projections[key] & other._projections[key]
                new_connect._projections[key].super = new_connect

            # leaf_connections
            key1 = set(self._leaf_connections.keys())
            key2 = set(other._leaf_connections.keys())
            unite_keys = key1.intersection(key2)
            for key in unite_keys:
                new_connect._leaf_connections[key] = self._leaf_connections[key] & other._leaf_connections[key]

        return new_connect

    def add_connection(self, con, name=None):
        assert isinstance(con, Projection)
        con.super = self
        key = con.key
        if not (con.pre in self.pre and con.post in self.post):
            raise ValueError(
                "the sub connection is not belong to this connection group (pre and post is not a member of the connected Assemblies)")

        con.set_name(name)
        if con.is_unit:
            self.__setattr__(name, con)
            # self._connections[key] = con
        else:
            self._projections[key] = con

    def add_leaf_connection(self, con):
        assert isinstance(con, Projection)
        con.super = self
        key = con.key
        if not (con.pre in self.pre and con.post in self.post):
            raise ValueError("the sub connection is not belong to this connection group (pre and post is not a member of the connected Assemblies)")
        con.set_name(None)
        self._leaf_connections[key] = con

    def del_connection(self):
        if self.super is not None:
            if self.key in self.super._connections:
                self.super._connections.pop(self.key)
            if self.key in self.super._projections:
                self.super._projections.pop(self.key)
            if self.key in self.super._leaf_connections:
                self.super._leaf_connections.pop(self.key)

            top_leaf = self.top._leaf_connections
            if self.key in top_leaf:
                top_leaf.pop(self.key)
            if self.super.is_empty():
                self.super.del_connection()
            self.super = None
        else:
            self.is_unit = False

    def expand_connection(self, to_level=-1):
        assert isinstance(to_level, int), ValueError("level is not int")
        if not bool(self._leaf_connections):
            self._leaf_connections[self.key] = self
        if self.is_unit:
            return self._leaf_connections.values()
        else:
            new_leaf_connections = dict()
            assb_connections = self._leaf_connections.values()
            while (len(assb_connections) > 0):
                if to_level >= 0 and to_level <= self.abstract_level:
                    break
                self.abstract_level += 1
                new_assb_connections = []
                for con in assb_connections:
                    if con.is_unit:
                        new_leaf_connections[con.key] = con
                    else:
                        pre_groups = con.pre.get_groups(recursive=False)
                        post_groups = con.post.get_groups(recursive=False)
                        for pre in pre_groups:
                            for post in post_groups:
                                new_con = Projection(pre, post)
                                # con.add_connection(new_con)
                                if new_con.is_unit:
                                    new_leaf_connections[new_con.key] = new_con
                                else:
                                    new_assb_connections.append(new_con)

                assb_connections = new_assb_connections

            for con in assb_connections:
                # 对于超过abstract_level的connection，都认为是leaf_connection
                new_leaf_connections[con.key] = con

            # self._leaf_connections = new_leaf_connections
            self._leaf_connections.clear()
            for key, con in new_leaf_connections.items():
                self.add_leaf_connection(con)

            return list(new_leaf_connections.values())

    def assembly_linked(self, assembly):
        if (assembly is self.pre) or (assembly is self.post):
            return True
        else:
            return False

    def replace_assembly(self, old_assembly, new_assembly):
        if old_assembly is self.pre:
            self.pre = new_assembly
        elif old_assembly is self.post:
            self.post = new_assembly
        else:
            raise ValueError("the old_assembly is not in the connnection")

    def add_super(self, assembly):
        assert isinstance(assembly, Assembly) or isinstance(assembly,
                                                            Projection), "the super is not Assembly or Projection"
        self._supers.append(assembly)

    def del_super(self, assembly):
        assert assembly in self._supers, "the assembly is not in supers"
        self._supers.remove(assembly)

    def get_connections(self, recursive=True):
        """
            Get the Connections in this assembly
        Args:
            recursive(bool): flag to decide if member connections of the member assemblies should be returned.

        Returns:
            List of Connections
        """
        if not recursive:
            return self._connections.values()
        else:
            connections = list(self._connections.values())
            for proj in self.sub_projections:
                connections.extend(proj.get_connections(recursive=False))
            return connections

    @property
    def top(self):
        if self.super is None:
            return self
        else:
            return self.super.top

    @property
    def sub_connections(self):
        if self.is_unit:
            raise ValueError("no sub_connections for unit connections")
        else:
            sub_connections = []
            sub_connections.extend(self._connections.values())
            return sub_connections

    @property
    def sub_projections(self):
        if self.is_unit:
            raise ValueError("no sub assb_connections for unit connections")
        else:
            return list(self._projections.values())

    @property
    def all_connections(self):
        if self.is_unit:
            return [self]
        else:
            unit_connections = []
            unit_connections.extend(self._connections.values())
            for con in self._projections.values():
                unit_connections.extend(con.all_connections)
            return unit_connections

    @property
    def all_projections(self):
        if self.is_unit:
            raise ValueError("no sub_connections for unit connections")
        else:
            assb_connections = []
            assb_connections.extend(self._projections.values())
            for con in self._projections.values():
                assb_connections.extend(con.all_projections)
            return assb_connections

    @property
    def leaf_connections(self):
        return list(self._leaf_connections.values())

    def get_str(self, level):

        # TODO: add str of subconnections
        level_space = "" + '-' * level
        repr_str = level_space + "|name:{}, type:{}, ".format(self.name, type(self).__name__)
        repr_str += "pre:{}, ".format(self.pre.name)
        repr_str += "post:{}\n ".format(self.post.name)
        level += 1
        for c in self._projections.values():
            repr_str += c.get_str(level)
        for c in self._connections.values():
            repr_str += c.get_str(level)
        return repr_str

    def train(self, mode=True):
        self.training = mode
        for p in self._projections.values():
            p.train(mode)
        for c in self._connections.values():
            c.training = mode

    def build(self, backend=None):
        self._backend = backend

        # clear previous builded connections
        if backend.builded == True:
            for super_asb in self._supers:
                for con in self._connections:
                    super_asb.del_connection(con)
        for sub_proj in self._projections.values():
            sub_proj.build(backend)
        # if not policy is not given then it does not have connection policy (no default policy)
        if not self._policies:
            return
        connection_inforamtion = None
        for p in self._policies:
            if connection_inforamtion is None:
                connection_inforamtion = p.generate_connection(self.pre, self.post)
            else:
                connection_inforamtion = connection_inforamtion & p.generate_connection(self.pre,
                                                                                        self.post)

        self._connections = connection_inforamtion._connections
        self._projections = connection_inforamtion._projections
        self._leaf_connections = connection_inforamtion._leaf_connections
        all_connections = dict()
        all_connections.update(self._connections)
        all_connections.update(self._leaf_connections)
        new_connections = dict()
        auto_con_num = 0
        for key, con in all_connections.items():
            assert con.is_unit
            if con.link_type is not None:
                link_type = con.link_type
                con_name = con.name
                con_kwargs = con.con_kwargs
                con_kwargs.update(self.ConnectionParameters)
                syn_kwargs = con.syn_kwargs
            else:
                assert self.link_type is not None
                link_type = self.link_type
                con_name = 'auto_con' + str(auto_con_num)
                auto_con_num += 1
                con_kwargs = self.ConnectionParameters
                syn_kwargs = None

            new_connections[key] = Connection(pre=con.pre, post=con.post,
                                              name=con_name, link_type=link_type, syn_kwargs=syn_kwargs, **con_kwargs)
            new_connections[key].add_super(self)
        self._connections = new_connections



class SynapseModel(ABC):
    '''
    op -> (return_name, operation_name, input_name1, input_name2...)
    '''

    #: A dictionary mapping synapse model names to `Model` objects
    synapse_models = dict()

    def __init__(self, conn, **kwargs):
        super(SynapseModel, self).__init__()
        self.name = 'none'
        self._syn_operations = []
        self._syn_variables = dict()
        self._syn_constant_variables = dict()
        self._syn_tau_variables = dict()
        if conn is not None:
            updated_input = conn.updated_input
            if updated_input:
                self.input_name = conn.pre_var_name + '[input][updated]'
            else:
                self.input_name = conn.pre_var_name + '[input]'

    @staticmethod
    def register(name, model):
        '''
        Register a synapse model. Registered synapse models can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. `'basic'`)
        model : `SynapseModel`
            The synapse model object, e.g. an `Basic_synapse`.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in SynapseModel.synapse_models:
            raise ValueError(('A synapse_model with the name "%s" has already been registered') % name)

        if not issubclass(model, SynapseModel):
            raise ValueError(('Given model of type %s does not seem to be a valid SynapseModel.' % str(type(model))))

        SynapseModel.synapse_models[name] = model
        model.name = name

    @staticmethod
    def apply_model(model_name):
        '''
        Parameters
        ----------
        model_name : str
        Returns
        -------
        '''
        model_name = model_name.lower()
        if model_name not in SynapseModel.synapse_models:
            raise ValueError(('Given synapse model name is not in the model list'))
        else:
            return SynapseModel.synapse_models[model_name]


class Connection(Projection):
    '''
    Base class for all kinds of connections, including full connection, sparse connection, conv connection,....
    Ten connection methods are provided, as shown below (key: class):
        'full', FullConnection
        'one_to_one_sparse', one_to_one_sparse
        'one_to_one', one_to_one_mask
        'conv', conv_connect
        'sparse_connection_sparse', sparse_connect_sparse
        'sparse_connection', sparse_connect_mask
        'random_connection_sparse', random_connect_sparse
        'random_connection', random_connect_mask
    Args:
        pre(Assembly): the assembly which needs to be connected.
        post(Assembly): the assembly which needs to connect the pre.
        link_type(str): the type for connection: full, sparse, conv...

    Attributes:
        pre_group(groups): the neuron group which need to be connected in the pre.
        post_group(groups): the neuron group which need to connect with pre_group neuron.
        _var_names(list): a list contain variable names.

    Methods:
        __new__: before build a new connection, do some checks.
        get_var_names: get variable names.
        register: register a connection class.
        build: add the connection variable, variable name and opperation to the backend.
        get_str:
        condition_check: check whether the pre_group.type is equal to the post_group.type, only if they are equal, return flag=Ture.
        connect: connect the preg with postg.
        get_weight_name: give a name for each connection weight.
        get_post_name: give a name for each post group.
        get_input_name: give a name for each input group.

    Examples:
        when building the network:
        self.connection1 = spaic.Connection(self.input, self.layer1, link_type='full')

        '''

    _connection_subclasses = dict()
    _class_label = '<con>'

    def __init__(self, pre: Assembly, post: Assembly, name=None,
                 link_type=('full', 'sparse_connection', 'conv', '...'),
                 syn_type=None, max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                 syn_kwargs=None, prefer_device=None, **kwargs):

        super(Connection, self).__init__(pre, post)
        # self.param_init = kwargs.get('param_init', None)
        self.con_kwargs = kwargs
        self.is_parameter = kwargs.get('is_parameter', True)
        self.is_sparse = kwargs.get('is_sparse', False)
        self.weight_quantization = kwargs.get('weight_quantization', False)
        if syn_kwargs is None:
            self.syn_kwargs = dict()
        else:
            self.syn_kwargs = syn_kwargs
        # assert pre._is_terminal
        # assert post._is_terminal
        self.prefer_device = prefer_device
        self.pre = pre  # .get_assemblies()
        self.post = post  # .get_assemblies()
        self.pre_var_name = pre_var_name
        self.post_var_name = post_var_name

        self.pre_num = pre.num
        self.post_num = post.num
        # self.link_num = self.pre_num * self.post_num
        self.shape = (self.post_num, self.pre_num)

        self.pre_groups = None  # pre.get_groups()
        self.post_groups = None  # post.get_groups()
        self.pre_assemblies = None
        self.post_assemblies = None
        self.connection_information = None
        self.mask_info: List[(tuple, tuple)] = list()  # (var_code, op_code)

        self.link_type = link_type
        self.max_delay = max_delay
        self.min_delay = kwargs.get('min_delay', 1.0)
        self.sparse_with_mask = sparse_with_mask
        self._var_names = list()
        self._var_dict = dict()
        self._supers = list()
        self._link_var_codes = list()
        # self._policies = policies
        self._variables = dict()
        self._operations = list()
        self._init_operations = list()
        # self._link_op_codes = list()
        self.w_init = None
        self.b_init = None
        self.w_init_param = dict()
        self.b_init_param = dict()

        self.parameters = kwargs
        # self.bias = kwargs.get('bias', False)

        # self.updated_input = kwargs.get('updated_input', True)
        self.weight_norm = kwargs.get('weight_norm', False)

        self.w_max = None
        self.w_min = None

        self.set_name(name)
        self.running_var = None
        self.running_mean = None
        self.decay = 0.9
        self._variables = dict()
        self._constant_variables = dict()

        self.flatten = kwargs.get('flatten', False)

        # if len(syn_type) == 0:
        #     raise ValueError('Please set the param syn_type, you can set it as \'basic\'')

        if isinstance(syn_type, list):
            self.synapse_type = syn_type
        else:
            self.synapse_type = [syn_type]
        self.synapse_name = []
        self.synapse_class = []

        for i in range(len(self.synapse_type)):
            if isinstance(self.synapse_type[i], str):
                self.synapse_class.append(SynapseModel.apply_model(self.synapse_type[i]))
                self.synapse_name.append(self.synapse_type[i])  # self.model -> self.model_name
                self.synapse = []
            else:
                raise ValueError("only support set synapse model with string")

    def __new__(cls, pre, post, name=None, link_type=('full', 'sparse_connection', 'conv', '...'),
                syn_type=None, max_delay=0, sparse_with_mask=False, pre_var_name='O', post_var_name='Isyn',
                syn_kwargs=None, **kwargs):
        if cls is not Connection:
            return super().__new__(cls)

        if link_type in cls._connection_subclasses:
            return cls._connection_subclasses[link_type](pre, post, name, link_type, syn_type,
                                                         max_delay, sparse_with_mask, pre_var_name, post_var_name,
                                                         syn_kwargs, **kwargs)

        else:
            raise ValueError("No connection type: %s in Connection classes" % link_type)

    def get_var_names(self):
        return self._var_names

    @staticmethod
    def register(name, connection_class):
        '''
        Register a connection class. Registered connection classes can be referred to
        # via their name.
        Parameters
        ----------
        name : str
            A short name for the state updater (e.g. `'full'`)
        connection_class :
            The subclass of Connection object, e.g. an `FullConnection`, 'ConvConnection'.
        '''

        # only deal with lower case names
        name = name.lower()
        if name in Connection._connection_subclasses:
            raise ValueError(('A connection class with the name "%s" has already been registered') % name)

        if not issubclass(connection_class, Connection):
            raise ValueError(
                ('Given model of type %s does not seem to be a valid ConnectionModel.' % str(type(connection_class))))

        Connection._connection_subclasses[name] = connection_class

    def assembly_linked(self, assembly):
        if (assembly is self.pre) or (assembly is self.post):
            return True
        else:
            return False

    def replace_assembly(self, old_assembly, new_assembly):
        if old_assembly is self.pre:
            self.pre = new_assembly
        elif old_assembly is self.post:
            self.post = new_assembly
        else:
            raise ValueError("the old_assembly is not in the connnection")

    def add_super(self, assembly):
        assert isinstance(assembly, Assembly) or isinstance(assembly, Projection), "the super is not Assembly"
        self._supers.append(assembly)

    def del_super(self, assembly):
        assert assembly in self._supers, "the assembly is not in supers"
        self._supers.remove(assembly)

    def set_id(self):
        if len(self._supers) == 0:
            self.id = self.name + self.__class__._class_label
        else:
            super_ids = []
            for super in self._supers:
                if super.id is not None:
                    super_ids.append(super.id)
                else:
                    super_ids.append(super.set_id())

            self.id = self.name + self.__class__._class_label
            if len(super_ids) == 1:
                self.id = super_ids[0] + '_' + self.id
            else:
                pre_id = '/'
                for prefix in super_ids:
                    pre_id += prefix + ','
                pre_id += '/'
                self.id = pre_id + '_' + self.id
        return self.id

    def get_str(self, level):

        level_space = "" + '-' * level
        repr_str = level_space + "|name:{}, type:{}, ".format(self.name, type(self).__name__)
        repr_str += "pre:{}, ".format(self.pre.name)
        repr_str += "post:{}\n ".format(self.post.name)
        # for c in self._connections.values():
        #     repr_str += c.get_str(level)
        return repr_str

    def set_delay(self, pre_group, post_group):
        # TODO: add to unit_connections information after it changed to List[dict]

        if self.max_delay > 0:
            # print("set delay")
            # pre_num = pre_group.num
            # post_num = post_group.num
            # shape = (post_num, pre_num)
            delay_input_name, delay_output_name = self.get_delay_input_output(pre_group, post_group)

            # add delay container
            delay_queue = self._backend.add_delay(delay_input_name, self.max_delay)
            delay_name = self.get_link_name(pre_group, post_group, 'delay')
            # ONLY FOR TEST  ===============================
            ini_delay = (self.max_delay-self.min_delay) * np.random.rand(*self.shape) + self.min_delay
            # ==============================================
            self.variable_to_backend(delay_name, self.shape, ini_delay, True)
            self.variable_to_backend(delay_output_name, (1, *self.shape), 0)

            # add delay index
            self._backend.register_standalone(Op(delay_output_name, delay_queue.select, [delay_name] ,owner=self))

            # add inital to transform initial delay_output
            # self.init_op_to_backend(delay_output_name, delay_queue.transform_delay_output,
            #                                [delay_input_name, delay_name])

        else:
            return

    def clamp_weight(self, weight):
        if not weight.is_sparse:
            if (self.w_max is not None) and (self.w_min is not None):
                self._backend.clamp_(weight, self.w_min, self.w_max)
            elif self.w_max is not None:
                self._backend.clamp_max_(weight, self.w_max)
            elif self.w_min is not None:
                self._backend.clamp_min_(weight, self.w_min)
    def quantize_weight(self, weight):
        return np.round(np.clip(weight + 2 ** 7, 0, 2 ** 8)) - 2 ** 7

    def _add_label(self, var_name: str):
        if '[pre]' in var_name:
            var_name = var_name.replace("[pre]", "")
            var_name = self.get_group_name(self.pre, var_name)
        elif '[link]' in var_name:
            var_name = var_name.replace("[link]", "")
            var_name = self.get_link_name(self.pre, self.post, var_name)
        elif '[post]' in var_name:
            var_name = var_name.replace("[post]", "")
            var_name = self.get_group_name(self.post, var_name)
        elif '[input]' in var_name:
            tmp_var_name = var_name.replace("[input]", "")
            var_name = self.get_input_name(self.pre, self.post)
            if tmp_var_name not in var_name:
                raise ValueError(" the var_name tagged as [input] is not pre_var_name of the connection")
        elif var_name[0] == '[' and var_name[-1] == ']':
            var_name = var_name
        else:
            var_name = self.get_name(var_name)
        return var_name

    def add_conn_label(self, var_name: str):
        if isinstance(var_name, str):
            if '[updated]' in var_name:
                var_name = var_name.replace("[updated]", "")
                name = self._add_label(var_name)
                return name + '[updated]'
            else:
                name = self._add_label(var_name)
                return name

        elif isinstance(var_name, list) or isinstance(var_name, tuple):
            var_names = []
            for name in var_name:
                if isinstance(name, str):
                    if '[updated]' in name:
                        name = name.replace("[updated]", "")
                        name = self._add_label(name)
                        name = name + '[updated]'
                    else:
                        name = self._add_label(name)
                    var_names.append(name)
            return var_names

    @property
    def dt(self):
        return self._backend.dt

    def decode_syn_op(self, syn_ops, synapse_name, op_len):
        if len(synapse_name) > 1:
            for i in range(1, op_len):
                pre_op_return_name = syn_ops[i-1][0]
                post_op_first_input = syn_ops[i][2]
                if '[updated]' in post_op_first_input:
                    post_op_first_input = post_op_first_input.replace('[updated]', '')
                if post_op_first_input == pre_op_return_name:
                    temp_name = synapse_name[i-1]+'_' + str(i-1)
                    syn_ops[i-1][0] = temp_name
                    syn_ops[i][2] = temp_name
        return syn_ops

    def decode_initializer(self, initial):
        init_name = initial.__class__.__name__
        init_param = initial.__dict__
        return init_name, init_param

    def build(self, backend):
        '''
        add the connection variable, variable name and operation to the backend.
        '''

        self._backend = backend
        # Add weight
        self.assigned_weight = False
        self.assigned_bias = False
        prefer_device=self.prefer_device if self.prefer_device != None else None

        self.set_delay(self.pre, self.post)


        for (key, value) in self._variables.items():
            var_name = self.add_conn_label(key)
            if isinstance(value, np.ndarray):
                if value.size > 1:
                    shape = value.shape
                else:
                    shape = ()
            elif hasattr(value, '__iter__'):
                value = np.array(value)
                if value.size > 1:
                    shape = value.shape
                else:
                    shape = ()
            else:
                shape = ()

            if 'weight' in key:
                if self.assigned_weight is False:
                    self.assigned_weight = True
                else:
                    raise ValueError("connection have multiple weight variable")
                if self.weight_quantization:
                    value = self.quantize_weight(value)
                self.weight = self.variable_to_backend(name=var_name, shape=self.shape, value=value, is_parameter=self.is_parameter,
                                         is_sparse=self.is_sparse, init=self.w_init, init_param=self.w_init_param, prefer_device=prefer_device)  # (var_name, shape, value, is_parameter, is_sparse, init)
            elif 'bias' in key:
                if self.assigned_bias is False:
                    self.assigned_bias = True
                else:
                    raise ValueError("connection have multiple bias variable")
                self.bias = self.variable_to_backend(name=var_name, shape=value.shape, value=value, is_parameter=self.is_parameter,
                                         init=self.b_init, init_param=self.b_init_param, prefer_device=prefer_device)
            elif hasattr(value, 'requires_grad'):
                self.variable_to_backend(var_name, shape=value.shape, value=value, is_parameter=value.requires_grad,
                                         prefer_device=prefer_device)
            else:
                self.variable_to_backend(var_name, shape, value=value, prefer_device=prefer_device)
            # self._var_names.append(var_name)
            # self._var_dict[var_name] = VariableAgent(backend, var_name)

        for (key, value) in self._constant_variables.items():
            var_name = self.add_conn_label(key)
            self.variable_to_backend(name=var_name, shape=None, value=value, is_constant=True, prefer_device=prefer_device)

        weight_name = self.get_link_name(self.pre, self.post, 'weight')
        if self.sparse_with_mask:
            mask = (self.weight != 0)
            mask_name = self.get_link_name(self.pre, self.post, 'mask')
            self.variable_to_backend(mask_name, self.shape, mask, prefer_device=prefer_device)
            self.init_op_to_backend(weight_name, self.mask_operation, [weight_name, mask_name], prefer_device)
            # del self.sparse_with_mask

        # del self.weight
        if (self.w_min is not None) or (self.w_max is not None):
            self.init_op_to_backend(None, self.clamp_weight, [weight_name], prefer_device)

        # TODO: 这里input 和 output的 name 默认没有加label,同时都加上conn的label是不是过于局限了？
        for op in self._operations:
            addcode_op = Op(owner=self)
            addcode_op.func_name = op[1]
            addcode_op.output = self.add_conn_label(op[0])
            if len(op) > 3:  # 为了解决历史的单一list格式的问题
                addcode_op.input = self.add_conn_label(op[2:])
            else:
                addcode_op.input = self.add_conn_label(op[2])
            addcode_op.place = prefer_device
            # if len(op) > 3:
            #     addcode_op[2] = addcode_op[2:]
            #     addcode_op = addcode_op[:3]
            backend.add_operation(addcode_op)

        for op in self._init_operations:
            addcode_op = []
            for ind, name in enumerate(op):
                if ind != 1:
                    addcode_op.append(self.add_conn_label(op[ind]))
                else:
                    addcode_op.append(op[ind])
            if len(op) > 3:
                addcode_op[2] = addcode_op[2:]
                addcode_op = addcode_op[:3]

            self.init_op_to_backend(addcode_op[0], addcode_op[1], addcode_op[2], self.prefer_device)

        # the initialzation of synapse with forward_build kwarg can be read
        if backend.forward_build:
            self.updated_input = True
        else:
            self.updated_input = False
        syn_ops = []
        for i in range(len(self.synapse_class)):
            if self.synapse_class[i] is not None:
                self.synapse.append(self.synapse_class[i](self, **self.syn_kwargs))

            for (key, value) in self.synapse[i]._syn_variables.items():
                key = self.add_conn_label(key)
                self.variable_to_backend(key, value.shape, value=value)

            for (key, value) in self.synapse[i]._syn_constant_variables.items():
                key = self.add_conn_label(key)
                self.variable_to_backend(key, shape=None, value=value, is_constant=True)
            dt = backend.dt
            for (key, value) in self.synapse[i]._syn_tau_variables.items():
                key = self.add_conn_label(key)
                tau_value = np.exp(-dt / value)
                # 暂时只考虑scalar值
                self.variable_to_backend(key, shape=[1, ], value=tau_value)
            for op in self.synapse[i]._syn_operations:
                syn_ops.append(op)

        #toDO: make sure synapse with multiple ops
        # if 'basic' in self.synapse_name:
        syn_ops = self.decode_syn_op(syn_ops, self.synapse_name, len(syn_ops))

        for sop in syn_ops:
            addcode_op = Op(owner=self)
            addcode_op.func_name = sop[1]
            addcode_op.output = self.add_conn_label(sop[0])
            if len(sop) > 3:  # 为了解决历史的单一list格式的问题
                addcode_op.input = self.add_conn_label(sop[2:])
            else:
                addcode_op.input = self.add_conn_label(sop[2])
            addcode_op.place = prefer_device
            backend.add_operation(addcode_op)

    @abstractmethod
    def condition_check(self, pre_group, post_group):
        NotImplemented()

    def mask_operation(self, weight, mask):
        return weight * mask

    def get_name(self, suffix_name: str):
        '''

        Args:
            pre_group(Assembly): the neuron group which needs to be connected
            post_group(Assembly): the neuron group which needs to connect with the pre_group
            suffix_name(str): represents the name of the object you want to retrieve, such as 'weight'
        Returns:
            name(str)
        '''

        name = self.id + ':' + '{' + suffix_name + '}'
        return name

    def get_link_name(self, pre_group: Assembly, post_group: Assembly, suffix_name: str):
        '''

        Args:
            pre_group(Assembly): the neuron group which needs to be connected
            post_group(Assembly): the neuron group which needs to connect with the pre_group
            suffix_name(str): represents the name of the object you want to retrieve, such as 'weight'
        Returns:
            name(str)
        '''

        name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + suffix_name + '}'
        return name

    def get_group_name(self, group: Assembly, suffix_name: str):
        '''
        Args:
            group(Assembly): the neuron group which needs to be connected
            suffix_name(str): represents the name of the object you want to retrieve, such as 'O'

        Returns:
            name(str)
        '''
        name = group.id + ':' + '{' + suffix_name + '}'
        return name

    # def get_post_name(self, post_group: Assembly, suffix_name: str):
    #     '''
    #     Args:
    #         post_group(Assembly): the neuron group which needs to connect with the pre_group
    #         suffix_name(str): represents the name of the object you want to retrieve, such as 'maxpool_kernel_size'
    #
    #     Returns:
    #         name(str)
    #     '''
    #     name = post_group.id + ':' + '{' + suffix_name + '}'
    #     return name

    # def get_mask_name(self, pre_group: Assembly, post_group: Assembly):
    #     name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{mask}'
    #     return name
    #
    # def get_delay_name(self, pre_group: Assembly, post_group: Assembly):
    #     name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{delay}'
    #     return name

    # 这两个特例应该要留着
    def get_delay_input_output(self, pre_group: Assembly, post_group: Assembly):
        input_name = pre_group.id + ':' + '{' + self.pre_var_name + '}'
        output_name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + self.pre_var_name + '}'
        return input_name, output_name

    def get_input_name(self, pre_group: Assembly, post_group: Assembly):
        '''
        Give a name for input group's output spikes,  the name consists of two parts: pre_group.id + ':' + '{0}
        Args:
            pre_group(Assembly): The neuron group which need to connect with post_group neuron.
            post_group(Assembly): the neuron group which needs to connect with the pre_group
        Returns:
            name(str)
        '''
        if self.max_delay > 0:
            name = self.id + ':' + post_group.id + '<-' + pre_group.id + ':' + '{' + self.pre_var_name + '}'
        else:
            name = pre_group.id + ':' + '{' + self.pre_var_name + '}'
        return name

    def get_target_output_name(self, output_group: Assembly):
        name = output_group.id + ':' + '{' + self.pre_var_name + '}'
        return name

    # def Chemistry_Isyn(self):
    #     """
    #     Chemistry current synapse
    #     I = tauP*I + WgtSum
    #     """
    #     I = self.get_post_name(self.post, 'I')
    #     WgtSum = self.get_post_name(self.post, 'WgtSum')
    #     tauP = self.get_post_name(self.post, 'tauP')
    #     # I = self.get_I_ele_name(self.post)
    #     # WgtSum = self.get_target_name(self.post)
    #     # tauP = self.post.id + ':' + '{tauP}'
    #     self._syn_variables[I] = 0
    #     self._syn_variables[WgtSum] = 0
    #     self._syn_tau_constant_variables[tauP] = self.tau_p
    #
    #     self._syn_operations.append([I, 'var_linear', tauP, I, WgtSum])

    # def Electrical_synapse(self):
    #     """
    #     Electrical synapse
    #     WgtSum = weight *（V(l-1) - V(l)）
    #     """
    #     V_post = self.get_post_name(self.post, 'V')
    #     V_pre = self.get_pre_name(self.post, 'V')
    #     Vtemp1_post = self.get_link_name(self.pre, self.post, 'Vtemp')
    #     I_post = self.get_post_name(self.post, 'I')
    #     weight = self.get_link_name(self.pre, self.post, 'weight')
    #     Vtemp1_pre = self.get_link_name(self.post, self.pre, 'Vtemp')
    #     I_pre = self.get_pre_name(self.post, 'I_ele')
    #     # V_post = self.get_V_name(self.post)
    #     # V_pre = self.get_V_name(self.pre)
    #     # Vtemp1_post = self.get_Vtemp_name(self.pre, self.post)
    #     # I_post = self.get_I_ele_name(self.post)
    #     # weight = self.get_weight_name(self.pre, self.post)
    #     # Vtemp1_pre = self.get_Vtemp_name(self.post, self.pre)
    #     # I_pre = self.get_I_ele_name(self.pre)
    #
    #     self._syn_variables[Vtemp1_post] = 0
    #     self._syn_variables[I_post] = 0
    #     self._syn_variables[Vtemp1_pre] = 0
    #     self._syn_variables[I_pre] = 0
    #     self._syn_operations.append([Vtemp1_post, 'minus', V_pre, V_post])
    #     self._syn_operations.append([I_post, 'var_mult', weight, Vtemp1_post + '[updated]'])
    #     self._syn_operations.append([Vtemp1_pre, 'minus', V_post, V_pre])
    #     self._syn_operations.append([I_pre, 'var_mult', weight, Vtemp1_pre + '[updated]'])


class ConnectPolicy():

    def __init__(self, level=-1):
        super(ConnectPolicy, self).__init__()
        self.level = level
        self.policies = [self, ]

    @abstractmethod
    def checked_connection(self, new_connection: Projection):
        NotImplementedError()

    def extend_policy(self, policies):
        self.policies.extend(policies)

    def __and__(self, other):
        self.extend_policy(other.policies)
        return self

    def generate_connection(self, pre: Assembly, post: Assembly):

        candidates = []
        for p in self.policies:
            candidates.append(p.checked_connection(Projection(pre, post)))
        res = candidates[-1]
        candidates.pop()
        for con in candidates:
            res = res & con
        return res